@@ -30,6 +30,7 @@
 ;;; Code:
 ;;
 (require 'elmo)
+(require 'elmo-signal)
 (require 'elmo-msgdb)
 
 (defvar elmo-filter-number-filename "number-list"
@@ -45,20 +46,52 @@
 					    name)
   (let (pair)
     (setq pair (elmo-parse-search-condition name))
-    (elmo-filter-folder-set-condition-internal folder
-					       (car pair))
+    (elmo-filter-folder-set-condition-internal folder (car pair))
     (if (string-match "^ */\\(.*\\)$" (cdr pair))
 	(elmo-filter-folder-set-target-internal
 	 folder
-	 (elmo-make-folder (elmo-match-string 1 (cdr pair))))
+	 (elmo-get-folder (elmo-match-string 1 (cdr pair))))
       (error "Folder syntax error `%s'" (elmo-folder-name-internal folder)))
     (elmo-filter-folder-set-require-msgdb-internal
      folder
      (elmo-folder-search-requires-msgdb-p
       (elmo-filter-folder-target-internal folder)
       (elmo-filter-folder-condition-internal folder)))
+    (elmo-filter-connect-signals
+     folder
+     (elmo-filter-folder-target-internal folder))
     folder))
 
+(defun elmo-filter-connect-signals (folder target)
+  (elmo-connect-signal
+   target 'flag-changing folder
+   (elmo-define-signal-handler (folder target number old-flags new-flags)
+     (elmo-filter-add-flag-count folder old-flags -1)
+     (elmo-filter-add-flag-count folder new-flags)
+     (elmo-emit-signal 'flag-changing folder number old-flags new-flags))
+   (elmo-define-signal-filter (folder target number)
+     (memq number (elmo-folder-list-messages folder nil t))))
+  (elmo-connect-signal
+   target 'flag-changed folder
+   (elmo-define-signal-handler (folder target numbers)
+     (let ((filterd (elmo-list-filter
+		     (elmo-folder-list-messages folder nil t)
+		     numbers)))
+       (when filterd
+	 (elmo-emit-signal 'flag-changed folder filterd)))))
+  (elmo-connect-signal
+   target 'cache-changed folder
+   (elmo-define-signal-handler (folder target number)
+     (elmo-emit-signal 'cache-changed folder number))
+   (elmo-define-signal-filter (folder target number)
+     (memq number (elmo-folder-list-messages folder nil t))))
+  (elmo-connect-signal
+   target 'update-overview folder
+   (elmo-define-signal-handler (folder target number)
+     (elmo-emit-signal 'update-overview folder number))
+   (elmo-define-signal-filter (folder target number)
+     (memq number (elmo-folder-list-messages folder nil t)))))
+
 (defun elmo-filter-number-list-load (dir)
   (elmo-object-load
    (expand-file-name elmo-filter-number-filename dir)))
@@ -86,12 +119,25 @@
 	  (setq flag-count (cons (cons flag delta) flag-count)))))
     (elmo-filter-folder-set-flag-count-internal folder flag-count)))
 
+(defun elmo-filter-add-flag-count (folder flags &optional delta)
+  (let ((flag-count (elmo-filter-folder-flag-count-internal folder))
+	(delta (or delta 1))
+	elem)
+    (dolist (flag flags)
+      (if (setq elem (assq flag flag-count))
+	  (setcdr elem (+ (cdr elem) delta))
+	(setq flag-count (cons (cons flag delta) flag-count))))
+    (elmo-filter-folder-set-flag-count-internal folder flag-count)))
+
 (defun elmo-filter-folder-flag-count (folder)
   (or (elmo-filter-folder-flag-count-internal folder)
       (elmo-filter-folder-countup-message-flags
        folder
        (elmo-folder-list-messages folder t t))))
 
+(defun elmo-filter-folder-copy-flag-count (flag-counts)
+  (mapcar (lambda (pair) (cons (car pair) (cdr pair))) flag-counts))
+
 (luna-define-method elmo-folder-open :after ((folder elmo-filter-folder)
 					     &optional load-msgdb)
   (when load-msgdb
@@ -109,11 +155,11 @@
 (luna-define-method elmo-folder-close-internal ((folder elmo-filter-folder))
   (elmo-folder-close-internal (elmo-filter-folder-target-internal folder)))
 
-(luna-define-method elmo-folder-close :after ((folder elmo-filter-folder))
+(luna-define-method elmo-folder-close ((folder elmo-filter-folder))
+  (elmo-generic-folder-close folder)
   (elmo-filter-folder-set-number-list-internal folder nil)
   (elmo-filter-folder-set-flag-count-internal folder nil)
-  (elmo-folder-set-msgdb-internal
-   (elmo-filter-folder-target-internal folder) nil))
+  (elmo-folder-close (elmo-filter-folder-target-internal folder)))
 
 (luna-define-method elmo-folder-commit ((folder elmo-filter-folder))
   (elmo-folder-commit (elmo-filter-folder-target-internal folder))
@@ -143,11 +189,14 @@
 (luna-define-method elmo-folder-newsgroups ((folder elmo-filter-folder))
   (elmo-folder-newsgroups (elmo-filter-folder-target-internal folder)))
 
-(luna-define-method elmo-find-fetch-strategy
-  ((folder elmo-filter-folder) entity &optional ignore-cache)
+(luna-define-method elmo-find-fetch-strategy ((folder elmo-filter-folder)
+					      number
+					      &optional
+					      ignore-cache
+					      require-entireness)
   (elmo-find-fetch-strategy
    (elmo-filter-folder-target-internal folder)
-   entity ignore-cache))
+   number ignore-cache require-entireness))
 
 (luna-define-method elmo-folder-get-primitive-list ((folder
 						     elmo-filter-folder))
@@ -174,15 +223,24 @@
 
 (luna-define-method elmo-message-fetch ((folder elmo-filter-folder)
 					number strategy
-					&optional section outbuf unseen)
-  (elmo-message-fetch
-   (elmo-filter-folder-target-internal folder)
-   number strategy section outbuf unseen))
+					&optional unseen section)
+  (elmo-message-fetch (elmo-filter-folder-target-internal folder)
+		      number strategy unseen section))
 
 (luna-define-method elmo-folder-delete-messages ((folder elmo-filter-folder)
 						 numbers)
-  (elmo-folder-delete-messages
-   (elmo-filter-folder-target-internal folder) numbers))
+  (let ((flag-count (elmo-filter-folder-copy-flag-count
+		     (elmo-filter-folder-flag-count-internal folder)))
+	(messages (copy-sequence
+		   (elmo-filter-folder-number-list-internal folder)))
+	success)
+    (elmo-folder-detach-messages folder numbers)
+    (unless (setq success
+		  (elmo-folder-delete-messages
+		   (elmo-filter-folder-target-internal folder) numbers))
+      (elmo-filter-folder-set-flag-count-internal folder flag-count)
+      (elmo-filter-folder-set-number-list-internal folder messages))
+    success))
 
 (luna-define-method elmo-folder-list-messages ((folder elmo-filter-folder)
 					       &optional visible-only in-msgdb)
@@ -294,6 +352,9 @@
 (luna-define-method elmo-folder-message-file-p ((folder elmo-filter-folder))
   (elmo-folder-message-file-p (elmo-filter-folder-target-internal folder)))
 
+(luna-define-method elmo-folder-local-p ((folder elmo-filter-folder))
+  (elmo-folder-local-p (elmo-filter-folder-target-internal folder)))
+
 (luna-define-method elmo-folder-plugged-p ((folder elmo-filter-folder))
   (elmo-folder-plugged-p (elmo-filter-folder-target-internal folder)))
 
@@ -307,16 +368,27 @@
   (elmo-message-file-name (elmo-filter-folder-target-internal folder)
 			  number))
 
-(luna-define-method elmo-message-flags ((folder elmo-filter-folder) number
-					&optional msgid)
+(luna-define-method elmo-message-flag-available-p ((folder
+						    elmo-filter-folder) number
+						    flag)
+  (elmo-message-flag-available-p
+   (elmo-filter-folder-target-internal folder)
+   number flag))
+
+(luna-define-method elmo-message-flags ((folder elmo-filter-folder) number)
   (elmo-message-flags (elmo-filter-folder-target-internal folder)
-		      number msgid))
+		      number))
 
 (luna-define-method elmo-message-set-cached ((folder elmo-filter-folder)
 					     number cached)
   (elmo-message-set-cached
    (elmo-filter-folder-target-internal folder) number cached))
 
+(luna-define-method elmo-message-number ((folder elmo-filter-folder)
+					 message-id)
+  (elmo-message-number (elmo-filter-folder-target-internal folder)
+		       message-id))
+
 (luna-define-method elmo-message-entity ((folder elmo-filter-folder) key)
   (elmo-message-entity (elmo-filter-folder-target-internal folder) key))
 
@@ -338,34 +410,36 @@
   (elmo-folder-close-flag-table (elmo-filter-folder-target-internal folder)))
 
 (luna-define-method elmo-folder-count-flags ((folder elmo-filter-folder))
-  (let* ((flag-count (elmo-filter-folder-flag-count folder))
-	 (new (or (cdr (assq 'new flag-count)) 0))
-	 (unread (or (cdr (assq 'unread flag-count)) 0))
-	 (answered(or (cdr (assq 'answered flag-count)) 0)))
-    (list new (- unread new) answered)))
+  (elmo-filter-folder-flag-count folder))
 
 (luna-define-method elmo-folder-set-flag ((folder elmo-filter-folder)
 					  numbers
 					  flag
 					  &optional is-local)
-  (elmo-filter-folder-countup-message-flags folder numbers -1)
   (elmo-folder-set-flag (elmo-filter-folder-target-internal folder)
-			numbers flag is-local)
-  (elmo-filter-folder-countup-message-flags folder numbers))
+			numbers flag is-local))
 
 (luna-define-method elmo-folder-unset-flag ((folder elmo-filter-folder)
 					    numbers
 					    flag
 					    &optional is-local)
-  (elmo-filter-folder-countup-message-flags folder numbers -1)
   (elmo-folder-unset-flag (elmo-filter-folder-target-internal folder)
-			  numbers flag is-local)
-  (elmo-filter-folder-countup-message-flags folder numbers))
+			  numbers flag is-local))
 
 (luna-define-method elmo-message-folder ((folder elmo-filter-folder)
 					 number)
   (elmo-message-folder (elmo-filter-folder-target-internal folder) number))
 
+(luna-define-method elmo-message-field ((folder elmo-filter-folder)
+					number field &optional type)
+  (elmo-message-field
+   (elmo-filter-folder-target-internal folder) number field type))
+
+(luna-define-method elmo-message-set-field ((folder elmo-filter-folder)
+					    number field value)
+  (elmo-message-set-field
+   (elmo-filter-folder-target-internal folder) number field value))
+
 (luna-define-method elmo-folder-clear ((folder elmo-filter-folder)
 				       &optional keep-killed)
   (unless keep-killed
@@ -407,9 +481,10 @@
 (luna-define-method elmo-folder-detach-messages ((folder elmo-filter-folder)
 						 numbers)
   (elmo-filter-folder-countup-message-flags folder numbers -1)
-  (elmo-list-delete numbers (elmo-filter-folder-number-list folder) #'delq)
-  (elmo-folder-detach-messages
-   (elmo-filter-folder-target-internal folder) numbers))
+  (elmo-filter-folder-set-number-list-internal
+   folder
+   (elmo-list-delete numbers (elmo-filter-folder-number-list folder) #'delq))
+  t)
 
 (luna-define-method elmo-folder-length ((folder elmo-filter-folder))
   (length (elmo-filter-folder-number-list-internal folder)))
