@@ -30,25 +30,44 @@
 (eval-when-compile (require 'cl))
 
 ;;; Code:
-(defcustom elmo-global-flag-list '(important)
+(defcustom elmo-global-flags '(important)
   "A list of flag symbol which is managed globally by the flag folder."
   :type '(repeat symbol)
   :group 'elmo)
 
+(defcustom elmo-local-flags '(unread any digest)
+  "A list of flag symbol which is not treated as global flag."
+  :type '(repeat symbol)
+  :group 'elmo)
+
 (defvar elmo-global-flag-folder-alist nil
   "Internal variable to hold global-flag-folder structures.")
 
+(eval-and-compile
+  (defconst elmo-flag-char-regexp "]!#$&'+,./0-9:;<=>?@A-Z[^_`a-z|}~-"))
+
+(defun elmo-flag-valid-p (flag)
+  (string-match (eval-when-compile
+		  (concat "^[" elmo-flag-char-regexp "]+$"))
+		(if (stringp flag) flag (symbol-name flag))))
+
 (eval-and-compile
   (luna-define-class elmo-flag-folder (elmo-localdir-folder)
-		     (flag minfo minfo-hash))
+		     (flag minfo minfo-hash max-number))
   (luna-define-internal-accessors 'elmo-flag-folder))
 
 (luna-define-method elmo-folder-initialize ((folder
 					     elmo-flag-folder)
 					    name)
-  (if (string-match "flag/\\([a-z]+\\)" name)
-      (setq name (match-string 1 name))
-    (setq name (symbol-name (car elmo-global-flag-list)))
+  (unless (string-match (eval-when-compile
+			  (concat "^flag\\(/\\(["
+				  elmo-flag-char-regexp
+				  "]+\\)\\)?$"))
+			name)
+    (error "Error in folder name `%s'" (elmo-folder-name-internal folder)))
+  (if (match-beginning 1)
+      (setq name (match-string 2 name))
+    (setq name (symbol-name (car elmo-global-flags)))
     (elmo-folder-set-name-internal
      folder
      (concat (elmo-folder-name-internal folder) "/" name)))
@@ -57,8 +76,8 @@
 	    msgdb-path)
 	(elmo-flag-folder-set-flag-internal folder flag)
 	(unless (elmo-global-flag-p flag)
-	  (setq elmo-global-flag-list
-		(nconc elmo-global-flag-list (list flag))))
+	  (setq elmo-global-flags
+		(nconc elmo-global-flags (list flag))))
 	;; must be AFTER set flag slot.
 	(setq msgdb-path (elmo-folder-msgdb-path folder))
 	(unless (file-directory-p msgdb-path)
@@ -69,6 +88,10 @@
 	(elmo-localdir-folder-set-directory-internal
 	 folder
 	 msgdb-path)
+	(if (file-exists-p (expand-file-name "max" msgdb-path))
+	    (elmo-flag-folder-set-max-number-internal
+	     folder
+	     (elmo-object-load (expand-file-name "max" msgdb-path))))
 	(if (file-exists-p (expand-file-name ".minfo" msgdb-path))
 	    (elmo-flag-folder-set-minfo-internal
 	     folder
@@ -101,7 +124,11 @@
 						elmo-flag-folder))
   (elmo-object-save
    (expand-file-name ".minfo" (elmo-folder-msgdb-path folder))
-   (elmo-flag-folder-minfo-internal folder)))
+   (elmo-flag-folder-minfo-internal folder))
+  (if (elmo-flag-folder-max-number-internal folder)
+      (elmo-object-save
+       (expand-file-name "max" (elmo-folder-msgdb-path folder))
+       (elmo-flag-folder-max-number-internal folder))))
 
 (luna-define-method elmo-folder-list-subfolders ((folder elmo-flag-folder)
 						 &optional one-level)
@@ -111,7 +138,7 @@
 	     (symbol-name (elmo-folder-type-internal folder))
 	     "/"
 	     (symbol-name flag)))
-	  elmo-global-flag-list))
+	  elmo-global-flags))
 
 (defun elmo-flag-folder-delete-message (folder number
 					       &optional keep-referrer)
@@ -126,7 +153,7 @@
 			     (elmo-flag-folder-minfo-hash-internal
 			      folder))
 	(unless keep-referrer
-	  (setq target-folder (elmo-make-folder (car pair)))
+	  (setq target-folder (elmo-get-folder (car pair)))
 	  (elmo-folder-open target-folder 'load-msgdb)
 	  ;; Unset the flag of the original folder.
 	  ;; (XXX Should the message-id checked?)
@@ -143,9 +170,9 @@
      (delq elem (elmo-flag-folder-minfo-internal folder))))
   t)
 
-(luna-define-method elmo-folder-delete-messages ((folder
-						  elmo-flag-folder)
-						 numbers)
+(luna-define-method elmo-folder-delete-messages-internal ((folder
+							   elmo-flag-folder)
+							  numbers)
   (dolist (number numbers)
     (elmo-flag-folder-delete-message folder number)
     (elmo-localdir-delete-message folder number))
@@ -196,12 +223,19 @@
 					       &optional flag number)
   (error "Cannot append to the flag folder"))
 
+(luna-define-method elmo-folder-unset-flag :before ((folder elmo-flag-folder)
+						    numbers
+						    flag
+						    &optional is-local)
+  (when (eq flag (elmo-flag-folder-flag-internal folder))
+    (error "Cannot unset flag `%s' in this folder" flag)))
+
 ;;; Utilities
 
 (defmacro elmo-flag-get-folder (flag)
   "Get the flag folder structure for FLAG."
-  `(when (memq ,flag elmo-global-flag-list)
-     (elmo-make-folder (concat  "'flag/" (symbol-name ,flag)))))
+  `(when (memq ,flag elmo-global-flags)
+     (elmo-get-folder (concat  "'flag/" (symbol-name ,flag)))))
 
 (defun elmo-flag-folder-referrer (folder number)
   "Return a list of referrer message information.
@@ -217,13 +251,13 @@ NUMBER is the number of the message."
 ;;; Global-Flag API
 (defun elmo-global-flag-p (flag)
   "Return non-nil when FLAG is global."
-  (memq flag elmo-global-flag-list))
+  (memq flag elmo-global-flags))
 
 (defun elmo-global-flags (fname number)
   "Return a list of global flags for the message.
 FNAME is the name string of the folder.
 NUMBER is the number of the message."
-  (let ((flag-list elmo-global-flag-list)
+  (let ((flag-list elmo-global-flags)
 	folder matches)
     (while flag-list
       (setq folder (elmo-flag-get-folder (car flag-list)))
@@ -241,11 +275,10 @@ FOLDER is the elmo folder structure.
 FLAG is the symbol of the flag."
   (when (elmo-global-flag-p flag)
     (let ((flag-folder (elmo-flag-get-folder flag))
-	  result entity)
+	  result number)
       (dolist (elem (elmo-flag-folder-minfo-internal flag-folder))
-	(if (setq entity (elmo-message-entity folder (nth 1 elem)))
-	    (setq result (cons (elmo-message-entity-number entity)
-			       result))))
+	(if (setq number (elmo-message-number folder (nth 1 elem)))
+	    (setq result (cons number result))))
       result)))
 
 ;;;
@@ -260,7 +293,13 @@ NUMBER is the message number."
   (dolist (flag flags)
     (elmo-global-flag-set flag folder number message-id)))
 
+(defun elmo-local-flag-p (flag)
+  "Return non-nil when flag is not appropriate for global flag."
+  (memq flag elmo-local-flags))
+
 (defsubst elmo-global-flag-set-internal (flag folder number message-id)
+  (when (elmo-local-flag-p flag)
+    (error "Cannot treat `%s' as global flag" flag))
   (when message-id
     (let ((flag-folder (elmo-flag-get-folder flag))
 	  cache new-file new-number elem)
@@ -283,10 +322,17 @@ NUMBER is the message number."
 			       (elmo-flag-folder-minfo-hash-internal
 				flag-folder)))
 	;; Append new element.
+	(elmo-flag-folder-set-max-number-internal
+	 flag-folder
+	 (+ (or (elmo-flag-folder-max-number-internal flag-folder)
+		;; This is the first time.
+		(car (elmo-folder-status flag-folder)))
+	    1))
 	(setq new-file
 	      (expand-file-name
 	       (int-to-string
-		(setq new-number (1+ (car (elmo-folder-status flag-folder)))))
+		(setq new-number
+		      (elmo-flag-folder-max-number-internal flag-folder)))
 	       (elmo-localdir-folder-directory-internal flag-folder)))
 	(with-temp-buffer
 	  (setq cache (and message-id (elmo-file-cache-get message-id)))
@@ -294,9 +340,8 @@ NUMBER is the message number."
 	      (elmo-copy-file (elmo-file-cache-path cache)
 			      new-file)
 	    (when (and folder number)
-	      (elmo-message-fetch folder number (elmo-make-fetch-strategy
-						 'entire)
-				  nil (current-buffer))
+	      (elmo-message-fetch folder number
+				  (elmo-make-fetch-strategy 'entire))
 	      (write-region-as-binary (point-min) (point-max) new-file nil
 				      'no-msg))))
 	(elmo-flag-folder-set-minfo-internal
@@ -342,7 +387,8 @@ If optional DELETE-IF-NONE is non-nil, delete message from flag folder when
 the message is not flagged in any folder.
 If DELETE-IF-NONE is a symbol `always',
 delete message without flagged in other folder."
-  (unless (eq (elmo-folder-type-internal folder) 'flag)
+  (unless (and (eq (elmo-folder-type-internal folder) 'flag)
+	       (eq (elmo-flag-folder-flag-internal folder) flag))
     (let ((flag-folder (elmo-flag-get-folder flag))
 	  elem key)
       (when flag-folder
@@ -373,7 +419,7 @@ NUMBERS is the message number list.
 If optional DELETE-IF-NONE is non-nil, delete message from flag folder when
 the message is not flagged in any folder."
   (unless (eq (elmo-folder-type-internal folder) 'flag)
-    (dolist (flag elmo-global-flag-list)
+    (dolist (flag elmo-global-flags)
       (dolist (number numbers)
 	(elmo-global-flag-detach flag folder number delete-if-none)))))
 
@@ -384,17 +430,33 @@ If FLAGS is `t', all global flags becomes candidates.
 If optional IGNORE-PRESERVED is non-nil, preserved flags
 \(answered, cached, new, unread\) are not included."
   (let ((result (copy-sequence (if (eq flags t)
-				   (setq flags elmo-global-flag-list)
+				   (setq flags elmo-global-flags)
 				 flags))))
     (while flags
       (unless (elmo-global-flag-p (car flags))
 	(setq result (delq (car flags) result)))
       (setq flags (cdr flags)))
     (when ignore-preserved
-      (dolist (flag '(answered cached new unread))
+      (dolist (flag elmo-preserved-flags)
 	(setq result (delq flag result))))
     result))
 
+(defun elmo-global-flags-initialize (&optional additional-flags)
+  (let ((dir (expand-file-name "flag" elmo-msgdb-directory)))
+    (setq elmo-global-flags
+	  (elmo-list-delete
+	   elmo-local-flags
+	   (elmo-uniq-list
+	    (append
+	     elmo-global-flags
+	     additional-flags
+	     (and (file-directory-p dir)
+		  (mapcar 'intern
+			  (elmo-list-delete
+			   '(".." ".")
+			   (directory-files dir))))))
+	   #'delq))))
+
 ;;; To migrate from global mark folder
 (defvar elmo-global-mark-filename "global-mark"
   "Obsolete variable. (Just for migration)")
@@ -404,7 +466,7 @@ If optional IGNORE-PRESERVED is non-nil, preserved flags
   (when (and (file-exists-p (expand-file-name elmo-global-mark-filename
 					      elmo-msgdb-directory))
 	     (elmo-global-flag-p 'important)
-	     (not (file-exists-p (elmo-folder-expand-msgdb-path
+	     (not (file-exists-p (elmo-folder-msgdb-path
 				  (elmo-flag-get-folder 'important)))))
     (elmo-global-mark-upgrade)))
 
@@ -427,6 +489,15 @@ If optional IGNORE-PRESERVED is non-nil, preserved flags
 	    (elmo-global-flag-set 'important nil nil (car elem))))))
     (message "Upgrading flag structure...done")))
 
+(luna-define-method elmo-folder-delete :around ((folder elmo-flag-folder))
+  (let ((flag (elmo-flag-folder-flag-internal folder)))
+    (when (luna-call-next-method)
+      (setq elmo-global-flags (delq flag elmo-global-flags))
+      (setq elmo-global-flag-folder-alist
+	    (delq (assq flag elmo-global-flag-folder-alist)
+		  elmo-global-flag-folder-alist))
+      t)))
+
 (require 'product)
 (product-provide (provide 'elmo-flag) (require 'elmo-version))
 
