@@ -167,7 +167,12 @@ REGEXP should have a grouping for namespace prefix.")
 (defconst elmo-imap4-flag-specs '((important "\\Flagged")
 				  (read "\\Seen")
 				  (unread "\\Seen" 'remove)
-				  (answered "\\Answered")))
+				  (answered "\\Answered")
+				  ;; draft-melnikov-imap-keywords-03.txt
+				  (forwarded "$Forwarded")
+				  (work "$Work")
+				  (personal "$Personal")
+				  (shouldreply "$ShouldReply")))
 
 ;; For debugging.
 (defvar elmo-imap4-debug nil
@@ -209,12 +214,12 @@ Debug information is inserted in the buffer \"*IMAP4 DEBUG*\"")
 
 (defsubst elmo-imap4-decode-folder-string (string)
   (if elmo-imap4-use-modified-utf7
-      (utf7-decode-string string 'imap)
+      (utf7-decode string 'imap)
     string))
 
 (defsubst elmo-imap4-encode-folder-string (string)
   (if elmo-imap4-use-modified-utf7
-      (utf7-encode-string string 'imap)
+      (utf7-encode string 'imap)
     string))
 
 ;;; Response
@@ -443,16 +448,15 @@ If response is not `OK' response, causes error with IMAP response text."
 	 (mime-elmo-imap-location-folder-internal location)
 	 (mime-elmo-imap-location-number-internal location)
 	 (mime-elmo-imap-location-strategy-internal location)
-	 section
-	 (current-buffer)
-	 'unseen)
+	 'unseen
+	 section)
 	(buffer-string))
-    (elmo-message-fetch
+    (elmo-message-fetch-string
      (mime-elmo-imap-location-folder-internal location)
      (mime-elmo-imap-location-number-internal location)
      (mime-elmo-imap-location-strategy-internal location)
-     section
-     nil 'unseen)))
+     'unseen
+     section)))
 
 
 (luna-define-method mime-imap-location-bodystructure
@@ -758,28 +762,50 @@ Returns response value if selecting folder succeed. "
     (t
      (member "\\*" (elmo-imap4-session-flags-internal session)))))
 
+(defun elmo-imap4-flag-to-imap-search-key (flag)
+  (case flag
+    (read "seen")
+    (unread "unseen")
+    (important "flagged")
+    (answered "answered")
+    (new "new")
+    (t (concat
+	"keyword "
+	(or (car (cdr (assq flag elmo-imap4-flag-specs)))
+	    (symbol-name flag))))))
+
+(defun elmo-imap4-flag-to-imap-criteria (flag)
+  (case flag
+    ((any digest)
+     (let ((criteria "flagged")
+	   (global-flags (delq 'important (elmo-get-global-flags t t))))
+       (dolist (flag (delete 'new
+			     (delete 'cached
+				     (copy-sequence
+				      (case flag
+					(any
+					 elmo-preserved-flags)
+					(digest
+					 elmo-digest-flags))))))
+	 (setq criteria (concat "or "
+				(elmo-imap4-flag-to-imap-search-key flag)
+				" "
+				criteria)))
+       (while global-flags
+	 (setq criteria (concat "or keyword "
+				(symbol-name (car global-flags))
+				" "
+				criteria))
+	 (setq global-flags (cdr global-flags)))
+       criteria))
+    (t
+     (elmo-imap4-flag-to-imap-search-key flag))))
+
 (defun elmo-imap4-folder-list-flagged (folder flag)
   "List flagged message numbers in the FOLDER.
 FLAG is one of the `unread', `read', `important', `answered', `any'."
   (let ((session (elmo-imap4-get-session folder))
-	(criteria (case flag
-		    (read "seen")
-		    (unread "unseen")
-		    (important "flagged")
-		    (answered "answered")
-		    (new "new")
-		    (any "or answered or unseen flagged")
-		    (digest "or unseen flagged")
-		    (t (concat "keyword " (capitalize (symbol-name flag)))))))
-    ;; Add search keywords
-    (when (or (eq flag 'digest)(eq flag 'any))
-      (let ((flags (delq 'important (elmo-get-global-flags t t))))
-	(while flags
-	  (setq criteria (concat "or keyword "
-				 (symbol-name (car flags))
-				 " "
-				 criteria))
-	  (setq flags (cdr flags)))))
+	(criteria (elmo-imap4-flag-to-imap-criteria flag)))
     (if (elmo-imap4-session-flag-available-p session flag)
 	(progn
 	  (elmo-imap4-session-select-mailbox
@@ -943,6 +969,7 @@ If CHOP-LENGTH is not specified, message set is not chopped."
 		  (forward-line -1)
 		  (not (elmo-imap4-parse-greeting)))
 	(accept-process-output process 1))
+      (erase-buffer)
       (set-process-filter process 'elmo-imap4-arrival-filter)
       (set-process-sentinel process 'elmo-imap4-sentinel)
 ;;;   (while (and (memq (process-status process) '(open run))
@@ -1809,9 +1836,7 @@ Return nil if no complete line has arrived."
 	(elmo-imap4-forward)
 	(nreverse body)))))
 
-(luna-define-method elmo-folder-initialize :around ((folder
-						     elmo-imap4-folder)
-						    name)
+(luna-define-method elmo-folder-initialize ((folder elmo-imap4-folder) name)
   (let ((default-user	elmo-imap4-default-user)
 	(default-server elmo-imap4-default-server)
 	(default-port	elmo-imap4-default-port)
@@ -1826,25 +1851,27 @@ Return nil if no complete line has arrived."
       ;; "hoge%imap.server@gateway".
       (setq default-user (elmo-match-string 1 default-server))
       (setq default-server (elmo-match-string 2 default-server)))
-    (setq name (luna-call-next-method))
     ;; mailbox
-    (setq parse (elmo-parse-token name ":"))
+    (setq parse (elmo-parse-token name ":@:!"))
     (elmo-imap4-folder-set-mailbox-internal folder
 					    (elmo-imap4-encode-folder-string
 					     (car parse)))
     ;; user
-    (setq parse (elmo-parse-prefixed-element ?: (cdr parse) "/"))
+    (setq parse (elmo-parse-prefixed-element ?: (cdr parse) "/@:!"
+					     "^[A-Za-z]+"))
     (elmo-net-folder-set-user-internal folder
 				       (if (eq (length (car parse)) 0)
 					   default-user
 					 (car parse)))
     ;; auth
-    (setq parse (elmo-parse-prefixed-element ?/ (cdr parse)))
+    (setq parse (elmo-parse-prefixed-element ?/ (cdr parse) "@:!"))
     (elmo-net-folder-set-auth-internal
      folder
      (if (eq (length (car parse)) 0)
 	 (or elmo-imap4-default-authenticate-type 'clear)
        (intern (car parse))))
+    ;; network
+    (elmo-net-parse-network folder (cdr parse))
     (unless (elmo-net-folder-server-internal folder)
       (elmo-net-folder-set-server-internal folder default-server))
     (unless (elmo-net-folder-port-internal folder)
@@ -1939,6 +1966,7 @@ Return nil if no complete line has arrived."
 	 (delim (or (cdr namespace-assoc)
 		 elmo-imap4-default-hierarchy-delimiter))
 	 ;; Append delimiter when root with namespace.
+	 (root-nodelim root)
 	 (root (if (and namespace-assoc
 			(match-end 1)
 			(string= (substring root (match-end 1))
@@ -1957,12 +1985,15 @@ Return nil if no complete line has arrived."
 		    (elmo-imap4-response-get-selectable-mailbox-list
 		     (elmo-imap4-send-command-wait
 		      session
-		      (list "list \"\" " (elmo-imap4-mailbox root)))))))
+		      (list "list \"\" " (elmo-imap4-mailbox
+					  root-nodelim)))))))
     (when (or (not (string= (elmo-net-folder-user-internal folder)
 			    elmo-imap4-default-user))
 	      (not (eq (elmo-net-folder-auth-internal folder)
 		       (or elmo-imap4-default-authenticate-type 'clear))))
-      (setq append-serv (concat ":" (elmo-net-folder-user-internal folder))))
+      (setq append-serv (concat ":"
+				(elmo-net-format-quoted
+				 (elmo-net-folder-user-internal folder) "/"))))
     (unless (eq (elmo-net-folder-auth-internal folder)
 		(or elmo-imap4-default-authenticate-type 'clear))
       (setq append-serv
@@ -2012,7 +2043,8 @@ Return nil if no complete line has arrived."
 				      fld))
 				  (cdr result)))
 		  folder (concat prefix
-				 (elmo-imap4-decode-folder-string folder)
+				 (elmo-net-format-quoted
+				  (elmo-imap4-decode-folder-string folder) ":")
 				 (and append-serv
 				      (eval append-serv)))
 		  ret (append ret (if has-child-p
@@ -2020,7 +2052,9 @@ Return nil if no complete line has arrived."
 				    (list folder)))))
 	  ret)
       (mapcar (lambda (fld)
-		(concat prefix (elmo-imap4-decode-folder-string fld)
+		(concat prefix
+			(elmo-net-format-quoted
+			 (elmo-imap4-decode-folder-string fld) ":")
 			(and append-serv
 			     (eval append-serv))))
 	      result))))
@@ -2052,13 +2086,14 @@ Return nil if no complete line has arrived."
 			       (elmo-folder-name-internal folder)))
       (let ((session (elmo-imap4-get-session folder)))
 	(when (elmo-imap4-folder-mailbox-internal folder)
-	  (when msgs (elmo-folder-delete-messages folder msgs))
+	  (when msgs (elmo-folder-delete-messages-internal folder msgs))
 	  (elmo-imap4-send-command-wait session "close")
 	  (elmo-imap4-send-command-wait
 	   session
 	   (list "delete "
 		 (elmo-imap4-mailbox
-		  (elmo-imap4-folder-mailbox-internal folder))))))
+		  (elmo-imap4-folder-mailbox-internal folder)))))
+	(elmo-imap4-session-set-current-mailbox-internal session nil))
       (elmo-msgdb-delete-path folder)
       t)))
 
@@ -2507,52 +2542,41 @@ If optional argument REMOVE is non-nil, remove FLAG."
 
 ;; elmo-folder-open-internal: do nothing.
 
-(luna-define-method elmo-find-fetch-strategy
-  ((folder elmo-imap4-folder) entity &optional ignore-cache)
-  (let ((number (elmo-message-entity-number entity))
-	cache-file size message-id)
-    (setq size (elmo-message-entity-field entity 'size))
-    (setq message-id (elmo-message-entity-field entity 'message-id))
-    (setq cache-file (elmo-file-cache-get message-id))
-    (if (or ignore-cache
-	    (null (elmo-file-cache-status cache-file)))
-	(if (and elmo-message-fetch-threshold
-		 (integerp size)
-		 (>= size elmo-message-fetch-threshold)
-		 (or (not elmo-message-fetch-confirm)
-		     (not (prog1 (y-or-n-p
+(luna-define-method elmo-find-fetch-strategy ((folder elmo-imap4-folder) number
+					      &optional
+					      ignore-cache
+					      require-entireness)
+  (let ((entity (elmo-message-entity folder number)))
+    (if (null entity)
+	(elmo-make-fetch-strategy 'entire)
+      (let* ((size (elmo-message-entity-field entity 'size))
+	     (message-id (elmo-message-entity-field entity 'message-id))
+	     (cache-file (elmo-file-cache-get message-id))
+	     (use-cache (and (not ignore-cache)
+			     (elmo-message-use-cache-p folder number)
+			     (if require-entireness
+				 (eq (elmo-file-cache-status cache-file)
+				     'entire)
+			       (elmo-file-cache-status cache-file)))))
+	(elmo-make-fetch-strategy
+	 (if use-cache
+	     (elmo-file-cache-status cache-file)
+	   (if (and (not require-entireness)
+		    elmo-message-fetch-threshold
+		    (integerp size)
+		    (>= size elmo-message-fetch-threshold)
+		    (or (not elmo-message-fetch-confirm)
+			(not (prog1
+				 (y-or-n-p
 				  (format
 				   "Fetch entire message at once? (%dbytes)"
 				   size))
-			    (message "")))))
-	    ;; Fetch message as imap message.
-	    (elmo-make-fetch-strategy 'section
-				      nil
-				      (elmo-message-use-cache-p
-				       folder number)
-				      (elmo-file-cache-path
-				       cache-file))
-	  ;; Don't use existing cache and fetch entire message at once.
-	  (elmo-make-fetch-strategy 'entire nil
-				    (elmo-message-use-cache-p
-				     folder number)
-				    (elmo-file-cache-path cache-file)))
-      ;; Cache found and use it.
-      (if (not ignore-cache)
-	  (if (eq (elmo-file-cache-status cache-file) 'section)
-	      ;; Fetch message with imap message.
-	      (elmo-make-fetch-strategy 'section
-					t
-					(elmo-message-use-cache-p
-					 folder number)
-					(elmo-file-cache-path
-					 cache-file))
-	    (elmo-make-fetch-strategy 'entire
-				      t
-				      (elmo-message-use-cache-p
-				       folder number)
-				      (elmo-file-cache-path
-				       cache-file)))))))
+			       (message "")))))
+	       'section
+	     'entire))
+	 use-cache
+	 (elmo-message-use-cache-p folder number)
+	 (elmo-file-cache-path cache-file))))))
 
 (luna-define-method elmo-folder-create-plugged ((folder elmo-imap4-folder))
   (elmo-imap4-send-command-wait
@@ -2561,6 +2585,18 @@ If optional argument REMOVE is non-nil, remove FLAG."
 	 (elmo-imap4-mailbox
 	  (elmo-imap4-folder-mailbox-internal folder)))))
 
+(defun elmo-imap4-flags-to-imap (flags)
+  "Convert FLAGS to the IMAP flag string."
+  (let ((imap-flag (if (not (memq 'unread flags)) "\\Seen")))
+    (dolist (flag flags)
+      (unless (memq flag '(new read unread cached))
+	(setq imap-flag
+	      (concat imap-flag
+		      (if imap-flag " ")
+		      (or (car (cdr (assq flag elmo-imap4-flag-specs)))
+			  (capitalize (symbol-name flag)))))))
+    imap-flag))
+
 (luna-define-method elmo-folder-append-buffer
   ((folder elmo-imap4-folder) &optional flags number)
   (if (elmo-folder-plugged-p folder)
@@ -2579,19 +2615,7 @@ If optional argument REMOVE is non-nil, remove FLAG."
 		    (elmo-imap4-mailbox (elmo-imap4-folder-mailbox-internal
 					 folder))
 		    (if (and flags (elmo-folder-use-flag-p folder))
-			(concat " ("
-				(mapconcat
-				 'identity
-				 (append
-				  (and (memq 'important flags)
-				       '("\\Flagged"))
-				  (and (not (memq 'unread flags))
-				       '("\\Seen"))
-				  (and (memq 'answered flags)
-				       '("\\Answered")))
-				 " ")
-				;; XX KEYWORD flags
-				") ")
+			(concat " (" (elmo-imap4-flags-to-imap flags) ") ")
 		      " () ")
 		    (elmo-imap4-buffer-literal send-buffer))))
 	  (kill-buffer send-buffer))
