@@ -45,6 +45,22 @@
 				    (unread ?S 'remove)
 				    (answered ?R)))
 
+(defcustom elmo-maildir-separator
+  (if (memq system-type
+	    '(windows-nt OS/2 emx ms-dos win32 w32 mswindows cygwin))
+      ?\- ?:)
+  "Character separating the id section from the flags section.
+According to the maildir specification, this should be a colon (?:),
+but some file systems don't support colons in filenames."
+  :type 'character
+  :group 'elmo)
+
+(defmacro elmo-maildir-adjust-separator (string)
+  `(if (= elmo-maildir-separator ?:)
+       ,string
+     (elmo-replace-in-string
+      ,string ":" (char-to-string elmo-maildir-separator))))
+
 ;;; ELMO Maildir folder
 (eval-and-compile
   (luna-define-class elmo-maildir-folder
@@ -97,11 +113,25 @@ LOCATION."
 	 (cur (directory-files cur-dir
 			       nil "^[^.].*$" t))
 	 unread-locations flagged-locations answered-locations
-	 sym locations flag-list)
+	 sym locations flag-list x-time y-time)
+    (setq cur (sort cur
+		    (lambda (x y)
+		      (setq x-time (elmo-get-last-modification-time
+				    (expand-file-name x cur-dir))
+			    y-time (elmo-get-last-modification-time
+				    (expand-file-name y cur-dir)))
+		      (cond
+		       ((< x-time y-time)
+			t)
+		       ((eq x-time y-time)
+			(< (elmo-maildir-sequence-number x)
+			   (elmo-maildir-sequence-number y)))))))
     (setq locations
 	  (mapcar
 	   (lambda (x)
-	     (if (string-match "^\\([^:]+\\):\\([^:]+\\)$" x)
+	     (if (string-match
+		  (elmo-maildir-adjust-separator "^\\([^:]+\\):\\([^:]+\\)$")
+		  x)
 		 (progn
 		   (setq sym (elmo-match-string 1 x)
 			 flag-list (string-to-char-list
@@ -223,7 +253,7 @@ LOCATION."
 	   'elmo-maildir-msgdb-create "Creating msgdb..."
 	   (/ (* i 100) len)))))
     (message "Creating msgdb...done")
-    (elmo-msgdb-sort-by-date new-msgdb)))
+    new-msgdb))
 
 (defun elmo-maildir-cleanup-temporal (dir)
   ;; Delete files in the tmp dir which are not accessed
@@ -260,14 +290,18 @@ LOCATION."
        (expand-file-name (car news) (expand-file-name "new" maildir))
        (expand-file-name (concat
 			  (car news)
-			  (unless (string-match ":2,[A-Z]*$" (car news))
-			    ":2,"))
+			  (unless (string-match
+				   (elmo-maildir-adjust-separator ":2,[A-Z]*$")
+				   (car news))
+			    (elmo-maildir-adjust-separator  ":2,")))
 			 (expand-file-name "cur" maildir)))
       (setq news (cdr news)))))
 
 (defun elmo-maildir-set-mark (filename mark)
   "Mark the FILENAME file in the maildir.  MARK is a character."
-  (if (string-match "^\\([^:]+:[12],\\)\\(.*\\)$" filename)
+  (if (string-match
+       (elmo-maildir-adjust-separator "^\\([^:]+:[12],\\)\\(.*\\)$")
+       filename)
       (let ((flaglist (string-to-char-list (elmo-match-string
 					    2 filename))))
 	(unless (memq mark flaglist)
@@ -277,12 +311,15 @@ LOCATION."
 			       (char-list-to-string flaglist)))))
     ;; Rescue no info file in maildir.
     (rename-file filename
-		 (concat filename ":2," (char-to-string mark))))
+		 (concat filename
+			 (elmo-maildir-adjust-separator ":2,")
+			 (char-to-string mark))))
   t)
 
 (defun elmo-maildir-delete-mark (filename mark)
   "Mark the FILENAME file in the maildir.  MARK is a character."
-  (if (string-match "^\\([^:]+:2,\\)\\(.*\\)$" filename)
+  (if (string-match (elmo-maildir-adjust-separator "^\\([^:]+:2,\\)\\(.*\\)$")
+		    filename)
       (let ((flaglist (string-to-char-list (elmo-match-string
 					    2 filename))))
 	(when (memq mark flaglist)
@@ -346,36 +383,24 @@ LOCATION."
 
 (defvar elmo-maildir-sequence-number-internal 0)
 
-(static-cond
- ((>= emacs-major-version 19)
-  (defun elmo-maildir-make-unique-string ()
-    "This function generates a string that can be used as a unique
+(defun elmo-maildir-sequence-number (file)
+  "Get `elmo-maildir' specific sequence number from FILE.
+Not that FILE is the name without directory."
+  ;; elmo-maildir specific.
+  (if (string-match "^.*_\\([0-9]+\\)\\..*" file)
+      (string-to-number (match-string 1 file))
+    -1))
+
+(defun elmo-maildir-make-unique-string ()
+  "This function generates a string that can be used as a unique
 file name for maildir directories."
-     (let ((cur-time (current-time)))
-       (format "%.0f.%d_%d.%s"
- 	      (+ (* (car cur-time)
-                    (float 65536)) (cadr cur-time))
-	      (emacs-pid)
-	      (incf elmo-maildir-sequence-number-internal)
-	      (system-name)))))
- ((eq emacs-major-version 18)
-  ;; A fake function for v18
-  (defun elmo-maildir-make-unique-string ()
-    "This function generates a string that can be used as a unique
-file name for maildir directories."
-    (unless (fboundp 'float-to-string)
-      (load-library "float"))
-    (let ((time (current-time)))
-      (format "%s%d.%d.%s"
-	      (substring
-	       (float-to-string
-		(f+ (f* (f (car time))
-			(f 65536))
-		    (f (cadr time))))
-	       0 5)
-	      (cadr time)
-	      (% (abs (random t)) 10000); dummy pid
-	      (system-name))))))
+  (let ((cur-time (current-time)))
+    (format "%.0f.%d_%d.%s"
+	    (+ (* (car cur-time)
+		  (float 65536)) (cadr cur-time))
+	    (emacs-pid)
+	    (incf elmo-maildir-sequence-number-internal)
+	    (system-name))))
 
 (defun elmo-maildir-temporal-filename (basedir)
   (let ((filename (expand-file-name
@@ -392,6 +417,23 @@ file name for maildir directories."
 	     basedir)))
     filename))
 
+(defun elmo-maildir-move-file (src dst)
+  (or (condition-case nil
+	  (progn
+	    ;; 1. Try add-link-to-file, then delete the original.
+	    ;;    This is safe on NFS.
+	    (add-name-to-file src dst)
+	    (ignore-errors
+	      ;; It's ok if the delete-file fails;
+	      ;; elmo-maildir-cleanup-temporal will catch it later.
+	      (delete-file src))
+	    t)
+	(error))
+      ;; 2. Even on systems with hardlinks, some filesystems (like AFS)
+      ;;    might not support them, so fall back on rename-file. This is
+      ;;    our best shot at atomic when add-name-to-file fails.
+      (rename-file src dst)))
+
 (luna-define-method elmo-folder-append-buffer ((folder elmo-maildir-folder)
 					       &optional flags number)
   (let ((basedir (elmo-maildir-folder-directory-internal folder))
@@ -405,8 +447,7 @@ file name for maildir directories."
 	    (copy-to-buffer dst-buf (point-min) (point-max)))
 	  (as-binary-output-file
 	   (write-region (point-min) (point-max) filename nil 'no-msg))
-	  ;; add link from new.
-	  (elmo-add-name-to-file
+	  (elmo-maildir-move-file
 	   filename
 	   (expand-file-name
 	    (concat "new/" (file-name-nondirectory filename))
@@ -455,19 +496,19 @@ file name for maildir directories."
 	    (succeeds numbers)
 	    filename flags id)
 	(dolist (number numbers)
-	  (setq flags (elmo-message-flags src-folder (car numbers))
+	  (setq flags (elmo-message-flags src-folder number)
 		filename (elmo-maildir-temporal-filename dir))
 	  (elmo-copy-file
 	   (elmo-message-file-name src-folder number)
 	   filename)
-	  (elmo-add-name-to-file
+	  (elmo-maildir-move-file
 	   filename
 	   (expand-file-name
 	    (concat "new/" (file-name-nondirectory filename))
 	    dir))
 	  ;; src folder's msgdb is loaded.
 	  (when (setq id (and src-msgdb-exists
-			      (elmo-message-field src-folder (car numbers)
+			      (elmo-message-field src-folder number
 						  'message-id)))
 	    (elmo-flag-table-set table id flags))
 	  (elmo-progress-notify 'elmo-folder-move-messages))
@@ -484,14 +525,18 @@ file name for maildir directories."
       (if (and file
 	       (file-writable-p file)
 	       (not (file-directory-p file)))
-	  (delete-file file)))))
+	  (delete-file file))))
+  t)
 
 (luna-define-method elmo-map-message-fetch ((folder elmo-maildir-folder)
 					    location strategy
 					    &optional section unseen)
   (let ((file (elmo-maildir-message-file-name folder location)))
     (when (file-exists-p file)
-      (insert-file-contents-as-binary file))))
+      (insert-file-contents-as-binary file)
+      (unless unseen
+	(elmo-map-folder-set-flag folder (list location) 'read))
+      t)))
 
 (luna-define-method elmo-folder-exists-p ((folder elmo-maildir-folder))
   (let ((basedir (elmo-maildir-folder-directory-internal folder)))
