@@ -28,8 +28,10 @@
 
 ;;; Code:
 ;;
+(eval-when-compile (require 'cl))
 
 (require 'elmo)
+(require 'elmo-signal)
 (require 'luna)
 
 (defvar elmo-multi-divide-number 100000
@@ -58,7 +60,7 @@
      folder
      (nconc (elmo-multi-folder-children-internal
 	     folder)
-	    (list (elmo-make-folder (car name)))))
+	    (list (elmo-get-folder (car name)))))
     (setq name (cdr name))
     (when (and (> (length name) 0)
 	       (eq (aref name 0) ?,))
@@ -66,8 +68,58 @@
   (elmo-multi-folder-set-divide-number-internal
    folder
    elmo-multi-divide-number)
+  (elmo-multi-connect-signals folder)
   folder)
 
+(defun elmo-multi-connect-signals (folder)
+  (elmo-connect-signal
+   nil 'flag-changing folder
+   (elmo-define-signal-handler (folder child number old-flags new-flags)
+     (elmo-emit-signal 'flag-changing folder
+		       (car (elmo-multi-map-numbers folder child (list number)))
+		       old-flags new-flags))
+   (elmo-define-signal-filter (folder sender)
+     (memq sender (elmo-multi-folder-children-internal folder))))
+  (elmo-connect-signal
+   nil 'flag-changed folder
+   (elmo-define-signal-handler (folder child numbers)
+     (elmo-emit-signal 'flag-changed folder
+		       (elmo-multi-map-numbers folder child numbers)))
+   (elmo-define-signal-filter (folder sender)
+     (memq sender (elmo-multi-folder-children-internal folder))))
+  (elmo-connect-signal
+   nil 'cache-changed folder
+   (elmo-define-signal-handler (folder child number)
+     (elmo-emit-signal
+      'flag-changed folder
+      (car (elmo-multi-map-numbers folder child (list number)))))
+   (elmo-define-signal-filter (folder sender)
+     (memq sender (elmo-multi-folder-children-internal folder))))
+  (elmo-connect-signal
+   nil 'update-overview folder
+   (elmo-define-signal-handler (folder child number)
+     (elmo-emit-signal
+      'update-overview folder
+      (car (elmo-multi-map-numbers folder child (list number)))))
+   (elmo-define-signal-filter (folder sender)
+     (memq sender (elmo-multi-folder-children-internal folder)))))
+
+(defun elmo-multi-map-numbers (folder child numbers)
+  (let ((multi (catch 'found
+		 (let ((children (elmo-multi-folder-children-internal folder))
+		       (index 0))
+		   (while children
+		     (setq index (1+ index))
+		     (when (eq (car children) child)
+		       (throw 'found index))
+		     (setq children (cdr children)))))))
+    (when multi
+      (let ((offset (* (elmo-multi-folder-divide-number-internal folder)
+		       multi)))
+      (mapcar (lambda (number) (+ offset number))
+	      numbers)))))
+
+
 (luna-define-method elmo-folder-open-internal ((folder elmo-multi-folder))
   (dolist (fld (elmo-multi-folder-children-internal folder))
     (elmo-folder-open-internal fld)))
@@ -80,9 +132,10 @@
   (dolist (fld (elmo-multi-folder-children-internal folder))
     (elmo-folder-close-internal fld)))
 
-(luna-define-method elmo-folder-close :after ((folder elmo-multi-folder))
+(luna-define-method elmo-folder-close ((folder elmo-multi-folder))
+  (elmo-generic-folder-close folder)
   (dolist (fld (elmo-multi-folder-children-internal folder))
-    (elmo-folder-set-msgdb-internal fld nil)))
+    (elmo-folder-close fld)))
 
 (luna-define-method elmo-folder-synchronize ((folder elmo-multi-folder)
 					     &optional
@@ -145,14 +198,26 @@
   (let ((pair (elmo-multi-real-folder-number folder number)))
     (elmo-message-set-cached (car pair) (cdr pair) cached)))
 
-(luna-define-method elmo-find-fetch-strategy
-  ((folder elmo-multi-folder) entity &optional ignore-cache)
-  (let ((pair (elmo-multi-real-folder-number
-	       folder
-	       (elmo-message-entity-number entity))))
-    (elmo-find-fetch-strategy
-     (car pair)
-     (elmo-message-entity (car pair) (cdr pair)) ignore-cache)))
+(luna-define-method elmo-find-fetch-strategy ((folder elmo-multi-folder)
+					      number
+					      &optional
+					      ignore-cache
+					      require-entireness)
+  (let ((pair (elmo-multi-real-folder-number folder number)))
+    (elmo-find-fetch-strategy (car pair)
+			      (cdr pair)
+			      ignore-cache
+			      require-entireness)))
+
+(luna-define-method elmo-message-number ((folder elmo-multi-folder)
+					 message-id)
+  (let ((children (elmo-multi-folder-children-internal folder))
+	match)
+    (while children
+      (when (setq match (elmo-message-number (car children) message-id))
+	(setq children nil))
+      (setq children (cdr children)))
+    match))
 
 (luna-define-method elmo-message-entity ((folder elmo-multi-folder) key)
   (cond
@@ -160,8 +225,9 @@
     (let* ((pair (elmo-multi-real-folder-number folder key))
 	   (entity (elmo-message-entity (car pair) (cdr pair))))
       (when entity
-	(elmo-message-entity-set-number (elmo-message-copy-entity entity)
-					key))))
+	(setq entity (elmo-message-copy-entity entity))
+	(elmo-message-entity-set-number entity key)
+	entity)))
    ((stringp key)
     (let ((children (elmo-multi-folder-children-internal folder))
 	  (cur-number 0)
@@ -186,14 +252,19 @@
    (elmo-message-entity-field entity 'references)))
 
 (luna-define-method elmo-message-field ((folder elmo-multi-folder)
-					number field)
+					number field &optional type)
   (let ((pair (elmo-multi-real-folder-number folder number)))
-    (elmo-message-field (car pair) (cdr pair) field)))
+    (elmo-message-field (car pair) (cdr pair) field type)))
 
-(luna-define-method elmo-message-flags ((folder elmo-multi-folder) number
-					&optional msgid)
+(luna-define-method elmo-message-flag-available-p ((folder
+						    elmo-multi-folder) number
+						    flag)
   (let ((pair (elmo-multi-real-folder-number folder number)))
-    (elmo-message-flags (car pair) (cdr pair) msgid)))
+    (elmo-message-flag-available-p (car pair) (cdr pair) flag)))
+
+(luna-define-method elmo-message-flags ((folder elmo-multi-folder) number)
+  (let ((pair (elmo-multi-real-folder-number folder number)))
+    (elmo-message-flags (car pair) (cdr pair))))
 
 (defun elmo-multi-split-numbers (folder numlist &optional as-is)
   (let ((numbers (sort numlist '<))
@@ -226,9 +297,9 @@
 
 (luna-define-method elmo-message-fetch ((folder elmo-multi-folder)
 					number strategy
-					&optional section outbuf unseen)
+					&optional unseen section)
   (let ((pair (elmo-multi-real-folder-number folder number)))
-    (elmo-message-fetch (car pair) (cdr pair) strategy section outbuf unseen)))
+    (elmo-message-fetch (car pair) (cdr pair) strategy unseen section)))
 
 (luna-define-method elmo-folder-delete-messages ((folder elmo-multi-folder)
 						 numbers)
@@ -248,26 +319,24 @@
   (elmo-multi-folder-diff folder))
 
 (defun elmo-multi-folder-diff (folder)
-  (let ((flds (elmo-multi-folder-children-internal folder))
-	(news 0)
+  (let ((news 0)
 	(unreads 0)
 	(alls 0)
-	no-unreads diff)
-    (while flds
-      (setq diff (elmo-folder-diff (car flds)))
-      (cond
-       ((consp (cdr diff)) ; (new unread all)
-	(setq news    (+ news (nth 0 diff))
-	      unreads (+ unreads (nth 1 diff))
-	      alls    (+ alls (nth 2 diff))))
-       (t
-	(setq no-unreads t)
-	(setq news    (+ news (car diff))
-	      alls    (+ alls (cdr diff)))))
-      (setq flds (cdr flds)))
-    (if no-unreads
-	(cons news alls)
-      (list news unreads alls))))
+	diff value)
+    (dolist (child (elmo-multi-folder-children-internal folder))
+      (setq diff (elmo-folder-diff child))
+      (setq news    (and news
+			 (setq value (elmo-diff-new diff))
+			 (+ news value))
+	    unreads (and unreads
+			 (setq value (elmo-diff-unread diff))
+			 (+ unreads value))
+	    alls    (and alls
+			 (setq value (elmo-diff-all diff))
+			 (+ alls value))))
+    (if unreads
+	(list news unreads alls)
+      (cons news alls))))
 
 (luna-define-method elmo-folder-list-messages
   ((folder elmo-multi-folder) &optional visible-only in-msgdb)
@@ -411,7 +480,8 @@
 					    flag
 					    &optional is-local)
   (dolist (pair (elmo-multi-make-folder-numbers-list folder numbers))
-    (elmo-folder-unset-flag (car pair) (cdr pair) flag is-local)))
+    (ignore-errors
+     (elmo-folder-unset-flag (car pair) (cdr pair) flag is-local))))
 
 (luna-define-method elmo-folder-list-flagged ((folder elmo-multi-folder)
 					      flag
@@ -443,16 +513,13 @@
     sum))
 
 (luna-define-method elmo-folder-count-flags ((folder elmo-multi-folder))
-  (let ((new 0)
-	(unreads 0)
-	(answered 0)
-	flags)
+  (let (flag-alist element)
     (dolist (child (elmo-multi-folder-children-internal folder))
-      (setq flags (elmo-folder-count-flags child))
-      (setq new (+ new (nth 0 flags)))
-      (setq unreads (+ unreads (nth 1 flags)))
-      (setq answered (+ answered (nth 2 flags))))
-    (list new unreads answered)))
+      (dolist (pair (elmo-folder-count-flags child))
+	(if (setq element (assq (car pair) flag-alist))
+	    (setcdr element (+ (cdr element) (cdr pair)))
+	  (setq flag-alist (cons pair flag-alist)))))
+    flag-alist))
 
 (require 'product)
 (product-provide (provide 'elmo-multi) (require 'elmo-version))
