@@ -56,6 +56,14 @@
 
 (defvar sasl-mechanisms)
 
+(defcustom elmo-network-session-idle-timeout nil
+  "Idle timeout of the network cache. Specified in seconds.
+If elapsed time since last access is larger than this value,
+cached session is not reused.
+If nil, network cache is reused."
+  :type '(choice number (const nil))
+  :group 'elmo)
+
 ;;; Code:
 ;;
 (eval-and-compile
@@ -66,7 +74,8 @@
 					      auth
 					      stream-type
 					      process
-					      greeting))
+					      greeting
+					      last-accessed))
   (luna-define-internal-accessors 'elmo-network-session))
 
 (luna-define-generic elmo-network-initialize-session (session)
@@ -166,16 +175,26 @@ if making session failed, returns nil."
     (setq pair (assoc (setq key (elmo-network-session-cache-key name folder))
 		      elmo-network-session-cache))
     (when (and pair
-	       (not (memq (process-status
-			   (elmo-network-session-process-internal
-			    (cdr pair)))
-			  '(open run))))
+	       (or (not (memq (process-status
+			       (elmo-network-session-process-internal
+				(cdr pair)))
+			      '(open run)))
+		   (and elmo-network-session-idle-timeout
+			(elmo-network-session-last-accessed-internal
+			 (cdr pair))
+			(elmo-time-expire
+			 (elmo-network-session-last-accessed-internal
+			  (cdr pair))
+			 elmo-network-session-idle-timeout))))
       (setq elmo-network-session-cache
 	    (delq pair elmo-network-session-cache))
       (elmo-network-close-session (cdr pair))
       (setq pair nil))
     (if pair
-	(cdr pair)			; connection cache exists.
+	(progn
+	  (elmo-network-session-set-last-accessed-internal
+	   (cdr pair) (current-time))
+	  (cdr pair))			; connection cache exists.
       (unless if-exists
 	(setq session
 	      (elmo-network-open-session
@@ -211,7 +230,9 @@ Returns a process object.  if making session failed, returns nil."
 			   :auth auth
 			   :stream-type stream-type
 			   :process nil
-			   :greeting nil))
+			   :greeting nil
+			   :last-accessed (current-time)
+			   ))
 	(buffer (format " *%s session for %s@%s:%d%s"
 			name
 			user
@@ -289,25 +310,34 @@ Returned value is searched from `elmo-network-stream-type-alist'."
       (setq alist (cdr alist)))
     spec))
 
-(luna-define-method elmo-folder-initialize ((folder
-					     elmo-net-folder)
-					    name)
+(defconst elmo-net-quote-chars "@:!")
+
+(defun elmo-net-format-quoted (string &optional extra-chars)
+  (if (string-match (concat "[" elmo-net-quote-chars extra-chars "]")
+		    string)
+      (elmo-quoted-token string)
+    string))
+
+(defun elmo-net-parse-network (folder network)
+  (let (parse)
+    ;; server
+    (setq parse (elmo-parse-prefixed-element ?@ network ":!"))
+    (when (> (length (car parse)) 0)
+      (elmo-net-folder-set-server-internal folder (car parse)))
+    ;; port
+    (setq parse (elmo-parse-prefixed-element ?: (cdr parse) "!"))
+    (when (> (length (car parse)) 0)
+      (elmo-net-folder-set-port-internal folder (string-to-int (car parse))))
+    ;; stream-type
+    (elmo-net-folder-set-stream-type-internal
+     folder
+     (assoc (cdr parse) elmo-network-stream-type-alist))))
+
+(luna-define-method elmo-folder-initialize ((folder elmo-net-folder) name)
   ;; user and auth should be set in subclass.
   (when (string-match "\\(@[^@:/!]+\\)?\\(:[0-9]+\\)?\\(!.*\\)?$" name)
-    (if (match-beginning 1)
-	(elmo-net-folder-set-server-internal
-	 folder
-	 (elmo-match-substring 1 name 1)))
-    (if (match-beginning 2)
-	(elmo-net-folder-set-port-internal
-	 folder
-	 (string-to-int (elmo-match-substring 2 name 1))))
-    (if (match-beginning 3)
-	(elmo-net-folder-set-stream-type-internal
-	 folder
-	 (assoc (elmo-match-string 3 name)
-		elmo-network-stream-type-alist)))
-    (substring name 0 (match-beginning 0))))
+    (elmo-net-parse-network folder (substring name (match-beginning 0))))
+  folder)
 
 (luna-define-method elmo-net-port-info ((folder elmo-net-folder))
   (list (elmo-net-folder-server-internal folder)
@@ -396,7 +426,7 @@ Returned value is searched from `elmo-network-stream-type-alist'."
 	 (elmo-delete-if
 	  (lambda (number) (memq number deleting))
 	  ;; current number-list.
-	  (elmo-msgdb-list-messages (elmo-folder-msgdb folder)))
+	  (elmo-folder-list-messages folder nil 'in-msgdb))
 	 ;; append appending messages
 	 (mapcar (lambda (x) (* -1 x))
 		 (elmo-dop-spool-folder-list-messages folder))))
@@ -414,11 +444,12 @@ Returned value is searched from `elmo-network-stream-type-alist'."
 						      flag)
   t)
 
-(luna-define-method elmo-folder-delete-messages ((folder elmo-net-folder)
-						 numbers)
-  (if (elmo-folder-plugged-p folder)
-      (elmo-folder-send folder 'elmo-folder-delete-messages-plugged numbers)
-    (elmo-folder-send folder 'elmo-folder-delete-messages-unplugged numbers)))
+(luna-define-method elmo-folder-delete-messages-internal ((folder
+							   elmo-net-folder)
+							  numbers)
+   (if (elmo-folder-plugged-p folder)
+       (elmo-folder-send folder 'elmo-folder-delete-messages-plugged numbers)
+     (elmo-folder-send folder 'elmo-folder-delete-messages-unplugged numbers)))
 
 (luna-define-method elmo-folder-delete-messages-unplugged ((folder
 							    elmo-net-folder)
