@@ -32,6 +32,7 @@
 
 ;;; Code:
 ;;
+(eval-when-compile (require 'cl))
 
 (require 'elmo-vars)
 (require 'elmo-util)
@@ -86,9 +87,7 @@ Debug information is inserted in the buffer \"*NNTP DEBUG*\"")
 		     (group temp-crosses reads))
   (luna-define-internal-accessors 'elmo-nntp-folder))
 
-(luna-define-method elmo-folder-initialize :around ((folder
-						     elmo-nntp-folder)
-						    name)
+(luna-define-method elmo-folder-initialize ((folder elmo-nntp-folder) name)
   (let ((elmo-network-stream-type-alist
 	 (if elmo-nntp-stream-type-alist
 	     (setq elmo-network-stream-type-alist
@@ -96,18 +95,20 @@ Debug information is inserted in the buffer \"*NNTP DEBUG*\"")
 			   elmo-network-stream-type-alist))
 	   elmo-network-stream-type-alist))
 	explicit-user parse)
-    (setq name (luna-call-next-method))
-    (setq parse (elmo-parse-token name ":"))
+    (setq parse (elmo-parse-token name ":@!"))
     (elmo-nntp-folder-set-group-internal folder
 					 (elmo-nntp-encode-group-string
 					  (car parse)))
     (setq explicit-user (eq ?: (string-to-char (cdr parse))))
-    (setq parse (elmo-parse-prefixed-element ?: (cdr parse)))
+    (setq parse (elmo-parse-prefixed-element ?: (cdr parse) "@:!"
+					     "^[A-Za-z]+"))
     (elmo-net-folder-set-user-internal folder
 				       (if (eq (length (car parse)) 0)
 					   (unless explicit-user
 					     elmo-nntp-default-user)
 					 (car parse)))
+    ;; network
+    (elmo-net-parse-network folder (cdr parse))
     (unless (elmo-net-folder-server-internal folder)
       (elmo-net-folder-set-server-internal folder
 					   elmo-nntp-default-server))
@@ -176,7 +177,8 @@ Don't cache if nil.")
 
 (defconst elmo-nntp-server-command-index '((xover . 0)
 					   (listgroup . 1)
-					   (list-active . 2)))
+					   (list-active . 2)
+					   (xhdr . 3)))
 
 (defmacro elmo-nntp-get-server-command (session)
   (` (assoc (cons (elmo-network-session-server-internal (, session))
@@ -505,7 +507,7 @@ Don't cache if nil.")
 			    (not (string= (elmo-nntp-folder-group-internal
 					   folder) "")))
 		       (concat " active"
-			       (format 
+			       (format
 				" %s.*"
 				(elmo-nntp-folder-group-internal folder))))))
 	  (if (elmo-nntp-read-response session t)
@@ -732,7 +734,7 @@ Don't cache if nil.")
   (let ((new-msgdb (elmo-make-msgdb))
 	ov-list message-id entity
 	ov-entity num
-	extras extra ext field field-index flags)
+	field field-index flags)
     (setq ov-list (elmo-nntp-parse-overview-string str))
     (while ov-list
       (setq ov-entity (car ov-list))
@@ -745,33 +747,32 @@ Don't cache if nil.")
       (setq num (string-to-int (aref ov-entity 0)))
       (when (or (null numlist)
 		(memq num numlist))
-	(setq extras elmo-msgdb-extra-fields
-	      extra nil)
-	(while extras
-	  (setq ext (downcase (car extras)))
-	  (when (setq field-index (cdr (assoc ext elmo-nntp-overview-index)))
-	    (when (> (length ov-entity) field-index)
-	      (setq field (aref ov-entity field-index))
-	      (when (eq field-index 8) ;; xref
-		(setq field (elmo-msgdb-remove-field-string field)))
-	      (setq extra (cons (cons ext field) extra))))
-	  (setq extras (cdr extras)))
 	(setq entity (elmo-msgdb-make-message-entity
 		      (elmo-msgdb-message-entity-handler new-msgdb)
 		      :message-id (aref ov-entity 4)
 		      :number     num
 		      :references (elmo-msgdb-get-last-message-id
 				    (aref ov-entity 5))
-		      :from       (elmo-mime-string (elmo-delete-char
-						     ?\"
-						     (or
-						      (aref ov-entity 2)
-						      elmo-no-from) 'uni))
-		      :subject    (elmo-mime-string (or (aref ov-entity 1)
-							elmo-no-subject))
+		      :from       (elmo-with-enable-multibyte
+				    (eword-decode-string
+				     (elmo-delete-char  ?\"
+							(or (aref ov-entity 2)
+							    elmo-no-from))))
+		      :subject    (or (elmo-with-enable-multibyte
+					(eword-decode-string
+					 (aref ov-entity 1)))
+				      elmo-no-subject)
 		      :date       (aref ov-entity 3)
-		      :size       (string-to-int (aref ov-entity 6))
-		      :extra      extra))
+		      :size       (string-to-int (aref ov-entity 6))))
+	(dolist (extra elmo-msgdb-extra-fields)
+	  (setq extra (downcase extra))
+	  (when (and (setq field-index
+			   (cdr (assoc extra elmo-nntp-overview-index)))
+		     (> (length ov-entity) field-index))
+	    (setq field (aref ov-entity field-index))
+	    (when (eq field-index 8) ;; xref
+	      (setq field (elmo-msgdb-remove-field-string field)))
+	    (elmo-message-entity-set-field entity (intern extra) field)))
 	(setq message-id (elmo-message-entity-field entity 'message-id)
 	      flags (elmo-flag-table-get flag-table message-id))
 	(elmo-global-flags-set flags folder num message-id)
@@ -959,8 +960,8 @@ Don't cache if nil.")
       (with-current-buffer (elmo-network-session-buffer session)
 	(std11-field-body "Newsgroups")))))
 
-(luna-define-method elmo-message-fetch-with-cache-process :around
-  ((folder elmo-nntp-folder) number strategy &optional section unread)
+(luna-define-method elmo-message-fetch :around
+  ((folder elmo-nntp-folder) number strategy &optional unread section)
   (when (luna-call-next-method)
     (elmo-nntp-setup-crosspost-buffer folder number)
     (unless unread
@@ -1186,10 +1187,16 @@ Returns a list of cons cells like (NUMBER . VALUE)"
 						      from-msgs)))
 	    result (sort result '<))))))
 
+(defun elmo-nntp-use-server-search-p (condition)
+  (if (vectorp condition)
+      (not (string= "body" (elmo-filter-key condition)))
+    (and (elmo-nntp-use-server-search-p (nth 1 condition))
+	 (elmo-nntp-use-server-search-p (nth 2 condition)))))
+
 (luna-define-method elmo-folder-search :around ((folder elmo-nntp-folder)
 						condition &optional from-msgs)
   (if (and (elmo-folder-plugged-p folder)
-	   (not (string= "body" (elmo-filter-key condition))))
+	   (elmo-nntp-use-server-search-p condition))
       (elmo-nntp-search-internal folder condition from-msgs)
     (luna-call-next-method)))
 
