@@ -30,6 +30,7 @@
 ;;
 
 (require 'elmo)
+(require 'elmo-signal)
 
 (defvar elmo-pipe-folder-copied-filename "copied"
   "Copied messages number set.")
@@ -52,8 +53,27 @@
     (elmo-pipe-folder-set-copy-internal folder
 					(string= ":"
 						 (elmo-match-string 2 name))))
+  (elmo-pipe-connect-signals folder (elmo-pipe-folder-dst-internal folder))
   folder)
 
+(defun elmo-pipe-connect-signals (folder destination)
+  (elmo-connect-signal
+   destination 'flag-changing folder
+   (elmo-define-signal-handler (folder dst number old-flags new-flags)
+     (elmo-emit-signal 'flag-changing folder number old-flags new-flags)))
+  (elmo-connect-signal
+   destination 'flag-changed folder
+   (elmo-define-signal-handler (folder dst numbers)
+     (elmo-emit-signal 'flag-changed folder numbers)))
+  (elmo-connect-signal
+   destination 'cache-changed folder
+   (elmo-define-signal-handler (folder dst number)
+     (elmo-emit-signal 'cache-changed folder number)))
+  (elmo-connect-signal
+   destination 'update-overview folder
+   (elmo-define-signal-handler (folder dst number)
+     (elmo-emit-signal 'update-overview folder number))))
+
 (luna-define-method elmo-folder-get-primitive-list ((folder elmo-pipe-folder))
   (nconc
    (elmo-folder-get-primitive-list (elmo-pipe-folder-src-internal folder))
@@ -78,9 +98,9 @@
 
 (luna-define-method elmo-message-fetch ((folder elmo-pipe-folder)
 					number strategy
-					&optional section outbuf unseen)
+					&optional unseen section)
   (elmo-message-fetch (elmo-pipe-folder-dst-internal folder)
-		      number strategy section outbuf unseen))
+		      number strategy unseen section))
 
 (luna-define-method elmo-folder-clear :after ((folder elmo-pipe-folder)
 					      &optional keep-killed)
@@ -92,6 +112,11 @@
   (elmo-folder-delete-messages (elmo-pipe-folder-dst-internal folder)
 			       numbers))
 
+(luna-define-method elmo-folder-detach-messages ((folder elmo-pipe-folder)
+						 numbers)
+  (elmo-folder-detach-messages (elmo-pipe-folder-dst-internal folder)
+			       numbers))
+
 (defvar elmo-pipe-drained-hook nil "A hook called when the pipe is flushed.")
 
 (defsubst elmo-pipe-folder-list-target-messages (src &optional ignore-list)
@@ -154,7 +179,11 @@
   (elmo-folder-open-internal (elmo-pipe-folder-dst-internal folder)))
 
 (luna-define-method elmo-folder-close-internal ((folder elmo-pipe-folder))
-  (elmo-folder-close-internal(elmo-pipe-folder-dst-internal folder)))
+  (elmo-folder-close-internal (elmo-pipe-folder-dst-internal folder)))
+
+(luna-define-method elmo-folder-close ((folder elmo-pipe-folder))
+  (elmo-generic-folder-close folder)
+  (elmo-folder-close (elmo-pipe-folder-dst-internal folder)))
 
 (luna-define-method elmo-folder-list-messages ((folder elmo-pipe-folder)
 					       &optional visible-only in-msgdb)
@@ -332,31 +361,44 @@
 (luna-define-method elmo-folder-length ((folder elmo-pipe-folder))
   (elmo-folder-length (elmo-pipe-folder-dst-internal folder)))
 
+(luna-define-method elmo-message-flag-available-p ((folder elmo-pipe-folder)
+						   number flag)
+  (elmo-message-flag-available-p
+   (elmo-pipe-folder-dst-internal folder)
+   number flag))
+
 (luna-define-method elmo-folder-count-flags ((folder elmo-pipe-folder))
   (elmo-folder-count-flags (elmo-pipe-folder-dst-internal folder)))
 
-(luna-define-method elmo-message-flags ((folder elmo-pipe-folder) number
-					&optional msgid)
-  (elmo-message-flags (elmo-pipe-folder-dst-internal folder) number msgid))
+(luna-define-method elmo-message-flags ((folder elmo-pipe-folder) number)
+  (elmo-message-flags (elmo-pipe-folder-dst-internal folder) number))
 
 (luna-define-method elmo-message-field ((folder elmo-pipe-folder)
-					number field)
+					number field &optional type)
   (elmo-message-field (elmo-pipe-folder-dst-internal folder)
 		      number
-		      field))
+		      field
+		      type))
 
 (luna-define-method elmo-message-set-cached ((folder elmo-pipe-folder)
 					     number cached)
   (elmo-message-set-cached (elmo-pipe-folder-dst-internal folder)
 			   number cached))
 
-(luna-define-method elmo-find-fetch-strategy
-  ((folder elmo-pipe-folder) entity &optional ignore-cache)
+(luna-define-method elmo-find-fetch-strategy ((folder elmo-pipe-folder)
+					      number
+					      &optional
+					      ignore-cache
+					      require-entireness)
   (elmo-find-fetch-strategy (elmo-pipe-folder-dst-internal folder)
-			    (elmo-message-entity
-			     (elmo-pipe-folder-dst-internal folder)
-			     (elmo-message-entity-number entity))
-			    ignore-cache))
+			    number
+			    ignore-cache
+			    require-entireness))
+
+(luna-define-method elmo-message-number ((folder elmo-pipe-folder)
+					 message-id)
+  (elmo-message-number (elmo-pipe-folder-dst-internal folder)
+		       message-id))
 
 (luna-define-method elmo-message-entity ((folder elmo-pipe-folder) key)
   (elmo-message-entity (elmo-pipe-folder-dst-internal folder) key))
