@@ -34,6 +34,7 @@
 (require 'shimbun)
 
 (eval-when-compile
+  (require 'cl)
   (defun-maybe shimbun-servers-list ()))
 
 (defcustom elmo-shimbun-check-interval 60
@@ -94,9 +95,7 @@ If it is the symbol `all', update overview for all shimbun folders."
   (luna-define-internal-accessors 'shimbun-elmo-mua))
 
 (luna-define-method shimbun-mua-search-id ((mua shimbun-elmo-mua) id)
-  (elmo-msgdb-message-entity (elmo-folder-msgdb
-			      (shimbun-elmo-mua-folder-internal mua))
-			     id))
+  (elmo-message-entity (shimbun-elmo-mua-folder-internal mua) id))
 
 (eval-and-compile
   (luna-define-class elmo-shimbun-folder
@@ -121,9 +120,7 @@ If it is the symbol `all', update overview for all shimbun folders."
 (defsubst elmo-shimbun-folder-shimbun-header (folder location)
   (let ((hash (elmo-shimbun-folder-header-hash-internal folder)))
     (or (and hash (elmo-get-hash-val location hash))
-	(let ((entity (elmo-msgdb-message-entity
-		       (elmo-folder-msgdb folder)
-		       location))
+	(let ((entity (elmo-message-entity folder location))
 	      (elmo-hash-minimum-size 63)
 	      header)
 	  (when entity
@@ -132,8 +129,7 @@ If it is the symbol `all', update overview for all shimbun folders."
 	      (elmo-shimbun-folder-set-header-hash-internal
 	       folder
 	       (setq hash (elmo-make-hash))))
-	    (elmo-set-hash-val (elmo-message-entity-field entity
-							  'message-id)
+	    (elmo-set-hash-val (elmo-message-entity-field entity 'message-id)
 			       header
 			       hash)
 	    header)))))
@@ -143,14 +139,6 @@ If it is the symbol `all', update overview for all shimbun folders."
     (+ (* (- (car now) (car time)) 65536)
        (- (nth 1 now) (nth 1 time)))))
 
-(defun elmo-shimbun-parse-time-string (string)
-  "Parse the time-string STRING and return its time as Emacs style."
-  (ignore-errors
-    (let ((x (timezone-fix-time string nil nil)))
-      (encode-time (aref x 5) (aref x 4) (aref x 3)
-		   (aref x 2) (aref x 1) (aref x 0)
-		   (aref x 6)))))
-
 (defsubst elmo-shimbun-headers-check-p (folder)
   (or (null (elmo-shimbun-folder-last-check-internal folder))
       (and (elmo-shimbun-folder-last-check-internal folder)
@@ -160,27 +148,24 @@ If it is the symbol `all', update overview for all shimbun folders."
 
 (defun elmo-shimbun-entity-to-header (entity)
   (let (message-id shimbun-id)
-    (if (setq message-id (elmo-message-entity-field
-			  entity 'x-original-id))
+    (if (setq message-id (elmo-message-entity-field entity 'x-original-id))
 	(setq shimbun-id (elmo-message-entity-field entity 'message-id))
       (setq message-id (elmo-message-entity-field entity 'message-id)
 	    shimbun-id nil))
-    (elmo-set-work-buf
-     (set-buffer-multibyte t)
-     (shimbun-make-header
-      (elmo-message-entity-number entity)
-      (shimbun-mime-encode-string
-       (elmo-message-entity-field entity 'subject 'decode))
-      (shimbun-mime-encode-string
-       (elmo-message-entity-field entity 'from 'decode))
-      (elmo-message-entity-field entity 'date)
-      message-id
-      (elmo-message-entity-field entity 'references)
-      0
-      0
-      (elmo-message-entity-field entity 'xref)
-      (and shimbun-id
-	   (list (cons "x-shimbun-id" shimbun-id)))))))
+    (elmo-with-enable-multibyte
+      (shimbun-create-header
+       (elmo-message-entity-number entity)
+       (elmo-message-entity-field entity 'subject)
+       (elmo-message-entity-field entity 'from)
+       (elmo-time-make-date-string
+	(elmo-message-entity-field entity 'date))
+       message-id
+       (elmo-message-entity-field entity 'references)
+       (elmo-message-entity-field entity 'size)
+       0
+       (elmo-message-entity-field entity 'xref)
+       (and shimbun-id
+	    (list (cons "x-shimbun-id" shimbun-id)))))))
 
 (defsubst elmo-shimbun-folder-header-hash-setup (folder headers)
   (let ((hash (or (elmo-shimbun-folder-header-hash-internal folder)
@@ -200,9 +185,7 @@ If it is the symbol `all', update overview for all shimbun folders."
 	  (delq nil
 		(mapcar
 		 (lambda (x)
-		   (unless (elmo-msgdb-message-entity
-			    (elmo-folder-msgdb folder)
-			    (shimbun-header-id x))
+		   (unless (elmo-message-entity folder (shimbun-header-id x))
 		     x))
 		 ;; This takes much time.
 		 (shimbun-headers
@@ -282,13 +265,13 @@ If it is the symbol `all', update overview for all shimbun folders."
    folder nil))
 
 (luna-define-method elmo-folder-plugged-p ((folder elmo-shimbun-folder))
-  (elmo-plugged-p
-   "shimbun"
-   (and (elmo-shimbun-folder-shimbun-internal folder)
-	(shimbun-server (elmo-shimbun-folder-shimbun-internal folder)))
-   nil nil
-   (and (elmo-shimbun-folder-shimbun-internal folder)
-	(shimbun-server (elmo-shimbun-folder-shimbun-internal folder)))))
+  (if (elmo-shimbun-folder-shimbun-internal folder)
+      (elmo-plugged-p
+       "shimbun"
+       (shimbun-server (elmo-shimbun-folder-shimbun-internal folder))
+       nil nil
+       (shimbun-server (elmo-shimbun-folder-shimbun-internal folder)))
+    t))
 
 (luna-define-method elmo-folder-set-plugged ((folder elmo-shimbun-folder)
 					     plugged &optional add)
@@ -374,53 +357,37 @@ If it is the symbol `all', update overview for all shimbun folders."
 	 percent))
       (setq numlist (cdr numlist)))
     (message "Creating msgdb...done")
-    (elmo-msgdb-sort-by-date new-msgdb)))
+    new-msgdb))
 
 (luna-define-method elmo-folder-message-file-p ((folder elmo-shimbun-folder))
   nil)
 
 (defsubst elmo-shimbun-update-overview (folder shimbun-id header)
-  (let ((entity (elmo-msgdb-message-entity (elmo-folder-msgdb folder)
-					   shimbun-id))
+  (let ((entity (elmo-message-entity folder shimbun-id))
 	(message-id (shimbun-header-id header))
 	references)
-    (unless (string= shimbun-id message-id)
-      (elmo-message-entity-set-field
-       entity 'x-original-id message-id)
-      (elmo-shimbun-header-set-extra-field
-       header "x-shimbun-id" shimbun-id)
-      (elmo-set-hash-val message-id
-			 entity
-			 (elmo-shimbun-folder-entity-hash folder))
-      (elmo-set-hash-val shimbun-id
-			 entity
-			 (elmo-shimbun-folder-entity-hash folder)))
-    (elmo-message-entity-set-field
-     entity
-     'from
-     (elmo-mime-string (shimbun-header-from header)))
-    (elmo-message-entity-set-field
-     entity
-     'subject
-     (elmo-mime-string (shimbun-header-subject header)))
-    (elmo-message-entity-set-field
-     entity
-     'date
-     (shimbun-header-date header))
-    (when (setq references
-		(or (elmo-msgdb-get-last-message-id
-		     (elmo-field-body "in-reply-to"))
-		    (elmo-msgdb-get-last-message-id
-		     (elmo-field-body "references"))))
-      (elmo-message-entity-set-field
-       entity
-       'references
-       (or (elmo-message-entity-field
-	    (elmo-get-hash-val
-	     references
-	     (elmo-shimbun-folder-entity-hash folder))
-	    'message-id)
-	   references)))))
+    (when (elmo-msgdb-update-entity
+	   (elmo-folder-msgdb folder)
+	   entity
+	   (nconc
+	    (unless (string= shimbun-id message-id)
+	      (elmo-shimbun-header-set-extra-field
+	       header "x-shimbun-id" shimbun-id)
+	      (elmo-set-hash-val message-id
+				 entity
+				 (elmo-shimbun-folder-entity-hash folder))
+	      (elmo-set-hash-val shimbun-id
+				 entity
+				 (elmo-shimbun-folder-entity-hash folder))
+	      (list (cons 'x-original-id message-id)))
+	    (list
+	     (cons 'from    (shimbun-header-from header 'no-encode))
+	     (cons 'subject (shimbun-header-subject header 'no-encode))
+	     (cons 'date    (shimbun-header-date header))
+	     (cons 'references
+		   (elmo-msgdb-get-references-from-buffer)))))
+      (elmo-emit-signal 'update-overview folder
+			(elmo-message-entity-number entity)))))
 
 (luna-define-method elmo-map-message-fetch ((folder elmo-shimbun-folder)
 					    location strategy
@@ -471,8 +438,7 @@ If it is the symbol `all', update overview for all shimbun folders."
 	       (when (and (elmo-message-entity-field ov 'xref)
 			  (if expire-days
 			      (< (elmo-shimbun-lapse-seconds
-				  (elmo-shimbun-parse-time-string
-				   (elmo-message-entity-field ov 'date)))
+				  (elmo-message-entity-field ov 'date))
 				 (* expire-days 86400 ; seconds per day
 				    ))
 			    t))
@@ -507,26 +473,34 @@ If it is the symbol `all', update overview for all shimbun folders."
 	       (setq folders
 		     (append folders
 			     (mapcar
-			      (lambda (fld) (concat prefix server "." fld))
+			      (lambda (group) (concat prefix server "." group))
 			      (shimbun-groups
-			       (shimbun-open server
-					     (let ((fld
-						    (elmo-make-folder
-						     (concat prefix server))))
-					       (luna-make-entity
-						'shimbun-elmo-mua
-						:folder fld))))))))
+			       (elmo-shimbun-folder-shimbun-internal
+				(elmo-get-folder (concat prefix server))))))))
 	     folders)))))
 
 (luna-define-method elmo-folder-exists-p ((folder elmo-shimbun-folder))
   (if (elmo-shimbun-folder-group-internal folder)
-      (progn
+      (if (fboundp 'shimbun-group-p)
+	  (shimbun-group-p (elmo-shimbun-folder-shimbun-internal folder)
+			   (elmo-shimbun-folder-group-internal folder))
 	(member
 	 (elmo-shimbun-folder-group-internal folder)
-	 (shimbun-groups (elmo-shimbun-folder-shimbun-internal
-			  folder))))
+	 (shimbun-groups (elmo-shimbun-folder-shimbun-internal folder))))
     t))
 
+(luna-define-method elmo-folder-delete-messages ((folder elmo-shimbun-folder)
+						 numbers)
+  (elmo-folder-kill-messages folder numbers)
+  t)
+
+(luna-define-method elmo-message-entity-parent ((folder elmo-shimbun-folder)
+						entity)
+  (let ((references (elmo-message-entity-field entity 'references)))
+    (and references
+	 (elmo-get-hash-val references
+			    (elmo-shimbun-folder-entity-hash folder)))))
+
 (require 'product)
 (product-provide (provide 'elmo-shimbun) (require 'elmo-version))
 
