@@ -57,16 +57,30 @@
   :prefix "elmo-"
   :group 'elmo)
 
+(defcustom elmo-digest-flags '(unread)
+  "Flags which are treated as `digest'."
+  :type '(repeat (symbol :tag "flag"))
+  :group 'elmo)
+
+(defcustom elmo-preserved-flags '(forwarded answered cached new unread)
+  "A list to define preserved flags.
+Flags in this list can be searched by `any' flag condition.
+If you want to treat a flag in this list as a `digest' flag,
+you have to add it to `elmo-digest-flags'.
+"
+  :type '(repeat (symbol :tag "flag"))
+  :group 'elmo)
+
 ;; Message Database
 
-(defcustom elmo-msgdb-default-type 'legacy
+(defcustom elmo-msgdb-default-type 'standard
   "*Default type of Message Database for ELMO."
   :type '(radio (const legacy)
 		(const standard))
   :group 'elmo
   :group 'elmo-setting)
 
-(defcustom elmo-msgdb-convert-type nil
+(defcustom elmo-msgdb-convert-type 'auto
   "*MODB conversion type."
   :type '(radio (const sync)
 		(const auto)
@@ -264,6 +278,13 @@ For disconnected operations.")
 (defvar elmo-strict-diff-folder-list nil
   "List of regexps of folder name which should be checked its diff strictly.")
 
+(defcustom elmo-msgdb-prefer-in-reply-to-for-parent nil
+  "*Non-nil to prefer In-Reply-To header for finding parent message on thread,
+rather than References header."
+  :type 'boolean
+  :group 'elmo
+  :group 'elmo-setting)
+
 (defcustom elmo-msgdb-extra-fields nil
   "Extra fields for msgdb."
   :type '(repeat string)
