@@ -35,6 +35,7 @@
 (require 'elmo-vars)
 (require 'elmo-util)
 (require 'elmo-msgdb)
+(require 'elmo-signal)
 
 (eval-when-compile (require 'cl))
 
@@ -59,7 +60,8 @@ Otherwise, entire fetching of the message is aborted without confirmation."
 
 (defcustom elmo-folder-update-threshold 500
   "Update threshold."
-  :type 'integer
+  :type '(choice (integer :tag "Number of messages")
+		 (const :tag "No limitation" nil))
   :group 'elmo)
 
 (defcustom elmo-folder-update-confirm t
@@ -67,6 +69,12 @@ Otherwise, entire fetching of the message is aborted without confirmation."
   :type 'boolean
   :group 'elmo)
 
+(defcustom elmo-msgdb-path-encode-threshold nil
+  "*Encode msgdb path if its length is longer than this value."
+  :type '(choice (const :tag "No encode" nil)
+		 number)
+  :group 'elmo)
+
 (defvar elmo-message-displaying nil
   "A global switch to indicate message is displaying or not.")
 
@@ -80,15 +88,34 @@ Otherwise, entire fetching of the message is aborted without confirmation."
 (elmo-define-error 'elmo-authenticate-error "Login failed" 'elmo-open-error)
 (elmo-define-error 'elmo-imap4-bye-error "IMAP4 session was terminated" 'elmo-open-error)
 
+;; Event declarations
+(elmo-define-signal flag-changing (number old-flags new-flags)
+  "Notify the changing flag of the messages with NUMBER.")
+
+(elmo-define-signal flag-changed (numbers)
+  "Notify the change flag of the messages with NUMBERS.")
+
+(elmo-define-signal cache-changed (number)
+  "Notify the change cache status of the message with NUMBER.")
+
+(elmo-define-signal update-overview (number)
+  "Notify update overview of the message with NUMBER.")
+
+(elmo-define-signal message-number-changed (old-number new-number)
+  "Notify change of message number within the folder.")
+
 ;; autoloads
 (eval-and-compile
+  (autoload 'md5 "md5")
   (autoload 'elmo-dop-queue-flush "elmo-dop")
   (autoload 'elmo-nntp-post "elmo-nntp")
   (autoload 'elmo-global-flag-p "elmo-flag")
+  (autoload 'elmo-local-flag-p "elmo-flag")
   (autoload 'elmo-global-flag-detach "elmo-flag")
   (autoload 'elmo-global-flag-detach-messages "elmo-flag")
   (autoload 'elmo-global-flag-set "elmo-flag")
   (autoload 'elmo-get-global-flags "elmo-flag")
+  (autoload 'elmo-global-flags-initialize "elmo-flag")
   (autoload 'elmo-global-mark-migrate "elmo-flag")
   (autoload 'elmo-folder-list-global-flag-messages "elmo-flag"))
 
@@ -122,6 +149,7 @@ If a folder name begins with PREFIX, use BACKEND."
 				     persistent   ; non-nil if persistent.
 				     process-duplicates  ; read or hide
 				     biff   ; folder for biff
+				     mime-charset ; charset for encode & decode
 				     ))
   (luna-define-internal-accessors 'elmo-folder))
 
@@ -134,16 +162,18 @@ If a folder name begins with PREFIX, use BACKEND."
   (` (luna-send (, folder) (, message) (, folder) (,@ args))))
 
 ;;;###autoload
-(defun elmo-make-folder (name &optional non-persistent)
+(defun elmo-make-folder (name &optional non-persistent mime-charset)
   "Make an ELMO folder structure specified by NAME.
-If optional argument NON-PERSISTENT is non-nil, the folder msgdb is not saved."
+If optional argument NON-PERSISTENT is non-nil, the folder msgdb is not saved.
+If optional argument MIME-CHARSET is specified, it is used for
+encode and decode a multibyte string."
   (let ((type (elmo-folder-type name))
 	prefix split class folder original)
     (setq original (elmo-string name))
     (if type
 	(progn
-	  (setq prefix (substring name 0 1))
-	  (setq name (substring name 1)))
+	  (setq prefix (elmo-string (substring name 0 1)))
+	  (setq name (elmo-string (substring name 1))))
       (setq type (intern (car (setq split (split-string name ":")))))
       (if (>= (length split) 2)
 	  (setq name (substring name (+ 1 (length (car split)))))
@@ -155,10 +185,18 @@ If optional argument NON-PERSISTENT is non-nil, the folder msgdb is not saved."
 				   :type type
 				   :prefix prefix
 				   :name original
-				   :persistent (not non-persistent)))
+				   :persistent (not non-persistent)
+				   :mime-charset mime-charset))
     (save-match-data
       (elmo-folder-send folder 'elmo-folder-initialize name))))
 
+(defvar elmo-get-folder-function nil)
+
+(defun elmo-get-folder (name)
+  (or (and elmo-get-folder-function
+	   (funcall elmo-get-folder-function name))
+      (elmo-make-folder name)))
+
 ;; Note that this function is for internal use only.
 (luna-define-generic elmo-folder-msgdb (folder)
   "Return the msgdb of FOLDER (on-demand loading).
@@ -314,12 +352,18 @@ Otherwise, all descendent folders are returned.")
   "Rename FOLDER to NEW-NAME (string).")
 
 (luna-define-generic elmo-folder-delete-messages (folder numbers)
-  "Delete messages.
+  "Delete messages with msgdb entity.
 FOLDER is the ELMO folder structure.
 NUMBERS is a list of message numbers to be deleted.
 It is not recommended to use this function other than internal use.
 Use `elmo-folder-move-messages' with dst-folder 'null instead.")
 
+(luna-define-generic elmo-folder-delete-messages-internal (folder numbers)
+  "Delete messages, but no delete msgdb entity.
+FOLDER is the ELMO folder structure.
+NUMBERS is a list of message numbers to be deleted.
+Override this method by each implement of `elmo-folder'.")
+
 (luna-define-generic elmo-folder-search (folder condition &optional numbers)
   "Search and return list of message numbers.
 FOLDER is the ELMO folder structure.
@@ -370,6 +414,9 @@ FLAG is a symbol which is one of the following:
   `all'       (remove all flags)
 If optional IS-LOCAL is non-nil, update only local (not server) status.")
 
+(luna-define-generic elmo-message-flag-available-p (folder number flag)
+  "Return non-nil when a message in the FOLDER with NUMBER treats FLAG.")
+
 (luna-define-generic elmo-folder-next-message-number (folder)
   "The next message number that will be assigned to a new message.
 FOLDER is the ELMO folder structure.")
@@ -378,8 +425,9 @@ FOLDER is the ELMO folder structure.")
 						       number)
   "Append current buffer as a new message.
 FOLDER is the destination folder (ELMO folder structure).
-FLAGS is the status of appended message (list of symbols).
-If it is nil, it is not that there is no flag and what is not defined is meant.
+FLAGS is the flag list for the appended message (list of symbols).
+If FLAGS contain `read', the message is appended as `not-unread'.
+If it is nil, the appended message will be treated as `new'.
 If optional argument NUMBER is specified, the new message number is set
 \(if possible\).
 Return nil on failure.")
@@ -455,22 +503,30 @@ Return newly created temporary directory name which contains temporary files.")
 (luna-define-generic elmo-message-file-p (folder number)
   "Return t if message in the FOLDER with NUMBER is a file.")
 
-(luna-define-generic elmo-message-flags (folder number &optional msgid)
+(luna-define-generic elmo-message-flags (folder number)
   "Return a list of flags.
 FOLDER is a ELMO folder structure.
-NUMBER is a number of the message.
-If optional argument MSGID is specified,
-the message with NUMBER checks whether it has MSGID.")
+NUMBER is a number of the message.")
 
-(luna-define-method elmo-message-flags ((folder elmo-folder) number
-					&optional msgid)
-  (if msgid
-      (let ((this-id (elmo-message-field folder number 'message-id)))
-	(and this-id
-	     (string= this-id msgid)
-	     (or (elmo-msgdb-flags (elmo-folder-msgdb folder) number)
-		 '(read))))
-    (elmo-msgdb-flags (elmo-folder-msgdb folder) number)))
+(defun elmo-message-flags-for-append (folder number &optional message-id)
+  "Return a list of flags for `elmo-folder-append-buffer'.
+FOLDER is a ELMO folder structure.
+NUMBER is a number of the message.
+If optional argument MESSAGES-ID is not specified, get it from current buffer."
+  (let ((this-id (elmo-message-field folder number 'message-id)))
+    (and this-id
+	 (string= this-id (or message-id
+			      (elmo-msgdb-get-message-id-from-buffer)))
+	 (or (elmo-message-flags folder number)
+	     ;; message exists, but no flag.
+	     '(read)))))
+
+(luna-define-method elmo-message-flag-available-p ((folder elmo-folder) number
+						   flag)
+  (elmo-msgdb-flag-available-p (elmo-folder-msgdb folder) flag))
+
+(luna-define-method elmo-message-flags ((folder elmo-folder) number)
+  (elmo-msgdb-flags (elmo-folder-msgdb folder) number))
 
 (defsubst elmo-message-flagged-p (folder number flag)
   "Return non-nil if the message is set FLAG.
@@ -483,97 +539,94 @@ NUMBER is a message number to test."
       (t
        (memq flag cur-flags)))))
 
-(luna-define-generic elmo-find-fetch-strategy
-  (folder entity &optional ignore-cache)
-;; Returns the message fetching strategy suitable for the message.
-;; FOLDER is the ELMO folder structure.
-;; ENTITY is the overview entity of the message in the folder.
-;; If optional argument IGNORE-CACHE is non-nil, cache is ignored.
-;; Returned value is a elmo-fetch-strategy object.
-;; If return value is nil, message should not be nil.
-  )
+(luna-define-generic elmo-find-fetch-strategy (folder number
+						      &optional
+						      ignore-cache
+						      require-entireness)
+  "Return the message fetching strategy suitable for the message with NUMBER.
+FOLDER is the ELMO folder structure.
+If optional argument IGNORE-CACHE is non-nil, existing cache is ignored.
+If second optional argument REQUIRE-ENTIRENESS is non-nil,
+ensure that entireness of the returned strategy is entire.
+Returned value is a elmo-fetch-strategy object.
+If return value is nil, message should not be nil.")
 
 (defmacro elmo-make-fetch-strategy (entireness
 				    &optional
 				    use-cache
 				    save-cache
 				    cache-path)
-;; Make elmo-message-fetching strategy.
-;; ENTIRENESS is 'entire or 'section.
-;; 'entire means fetch message entirely at once.
-;; 'section means fetch message section by section.
-;; If optional USE-CACHE is non-nil, existing cache is used and otherwise,
-;; existing cache is thrown away.
-;; If SAVE-CACHE is non-nil, fetched message is saved.
-;; CACHE-PATH is the cache path to be used as a message cache file.
-  (` (vector (, entireness)
-	     (, use-cache) (, save-cache) (, cache-path))))
+  "Make elmo-message-fetching strategy.
+ENTIRENESS is 'entire or 'section.
+'entire means fetch message entirely at once.
+'section means fetch message section by section.
+If optional USE-CACHE is non-nil, existing cache is used and otherwise,
+existing cache is thrown away.
+If SAVE-CACHE is non-nil, fetched message is saved.
+CACHE-PATH is the cache path to be used as a message cache file."
+  `(vector ,entireness ,use-cache ,save-cache ,cache-path))
 
 (defmacro elmo-fetch-strategy-entireness (strategy)
-  ;; Return entireness of STRATEGY.
-  (` (aref (, strategy) 0)))
+  "Return entireness of STRATEGY."
+  `(aref ,strategy 0))
 
 (defmacro elmo-fetch-strategy-use-cache (strategy)
-  ;; Return use-cache of STRATEGY.
-  (` (aref (, strategy) 1)))
+  "Return use-cache of STRATEGY."
+  `(aref ,strategy 1))
 
 (defmacro elmo-fetch-strategy-save-cache (strategy)
-  ;; Return save-cache of STRATEGY.
-  (` (aref (, strategy) 2)))
+  "Return save-cache of STRATEGY."
+  `(aref ,strategy 2))
 
 (defmacro elmo-fetch-strategy-cache-path (strategy)
-  ;;  Return cache-path of STRATEGY.
-  (` (aref (, strategy) 3)))
+  "Return cache-path of STRATEGY."
+  `(aref ,strategy 3))
 
-(luna-define-method elmo-find-fetch-strategy
-  ((folder elmo-folder) entity &optional ignore-cache)
-  (let (cache-file size message-id number)
-    (setq size (elmo-message-entity-field entity 'size))
-    (setq message-id (elmo-message-entity-field entity 'message-id))
-    (setq number (elmo-message-entity-number entity))
-    (setq cache-file (elmo-file-cache-get message-id))
-    (setq ignore-cache (or ignore-cache
-			   (null (elmo-message-use-cache-p folder number))))
-    (if (or ignore-cache
-	    (null (elmo-file-cache-status cache-file)))
-	;; No cache or ignore-cache.
-	(if (and (not (elmo-folder-local-p folder))
-		 elmo-message-fetch-threshold
-		 (integerp size)
-		 (>= size elmo-message-fetch-threshold)
-		 (or (not elmo-message-fetch-confirm)
-		     (not (prog1 (y-or-n-p
-				  (format "Fetch entire message(%dbytes)? "
-					  size))
-			    (message "")))))
-	    ;; Don't fetch message at all.
-	    nil
-	  ;; Don't use existing cache and fetch entire message at once.
-	  (elmo-make-fetch-strategy
-	   'entire nil
-	   (elmo-message-use-cache-p folder number)
-	   (elmo-file-cache-path cache-file)))
-      ;; Cache exists.
-      (if (not ignore-cache)
-	  (elmo-make-fetch-strategy
-	   'entire
-	   ;; ...But ignore current section cache and re-fetch
-	   ;; if section cache.
-	   (not (eq (elmo-file-cache-status cache-file) 'section))
-	   ;; Save cache.
-	   (elmo-message-use-cache-p folder number)
-	   (elmo-file-cache-path cache-file))))))
+(luna-define-method elmo-find-fetch-strategy ((folder elmo-folder) number
+					      &optional
+					      ignore-cache
+					      require-entireness)
+  (let ((entity (elmo-message-entity folder number)))
+    (if (null entity)
+	(elmo-make-fetch-strategy 'entire)
+      (let* ((size (elmo-message-entity-field entity 'size))
+	     (message-id (elmo-message-entity-field entity 'message-id))
+	     (cache-file (elmo-file-cache-get message-id))
+	     (use-cache (elmo-message-use-cache-p folder number)))
+	(if (and (not ignore-cache)
+		 use-cache
+		 (eq (elmo-file-cache-status cache-file) 'entire))
+	    ;; Cache exists and use it.
+	    (elmo-make-fetch-strategy
+	     'entire
+	     t				; Use cache.
+	     use-cache			; Save cache.
+	     (elmo-file-cache-path cache-file))
+	  ;; No cache or ignore-cache.
+	  (if (and (not (elmo-folder-local-p folder))
+		   (not require-entireness)
+		   elmo-message-fetch-threshold
+		   (integerp size)
+		   (>= size elmo-message-fetch-threshold)
+		   (or (not elmo-message-fetch-confirm)
+		       (not (prog1
+				(y-or-n-p
+				 (format "Fetch entire message(%dbytes)? "
+					 size))
+			      (message "")))))
+	      ;; Don't fetch message at all.
+	      nil
+	    ;; Don't use existing cache and fetch entire message at once.
+	    (elmo-make-fetch-strategy
+	     'entire
+	     nil			; Don't use cache.
+	     use-cache			; Save cache.
+	     (elmo-file-cache-path cache-file))))))))
 
 (luna-define-method elmo-folder-list-messages-internal
   ((folder elmo-folder) &optional visible-only)
   t)
 
-(defun elmo-folder-encache (folder numbers &optional unread)
-  "Encache messages in the FOLDER with NUMBERS.
-If UNREAD is non-nil, messages are not flaged as read."
-  (dolist (number numbers)
-    (elmo-message-encache folder number unread)))
-
 (luna-define-generic elmo-message-encache (folder number &optional read)
   "Encache message in the FOLDER with NUMBER.
 If READ is non-nil, message is flaged as read.")
@@ -581,15 +634,16 @@ If READ is non-nil, message is flaged as read.")
 (luna-define-method elmo-message-encache ((folder elmo-folder) number
 					  &optional read)
   (let (path)
-    (elmo-message-fetch
-     folder number
-     (elmo-make-fetch-strategy 'entire
-			       nil ;use-cache
-			       t   ;save-cache
-			       (setq path (elmo-file-cache-get-path
-					   (elmo-message-field
-					    folder number 'message-id))))
-     nil nil (not read))
+    (with-temp-buffer
+      (elmo-message-fetch
+       folder number
+       (elmo-make-fetch-strategy 'entire
+				 nil ;use-cache
+				 t   ;save-cache
+				 (setq path (elmo-file-cache-get-path
+					     (elmo-message-field
+					      folder number 'message-id))))
+       (not read)))
     path))
 
 (luna-define-generic elmo-message-fetch-bodystructure (folder number strategy)
@@ -597,32 +651,15 @@ If READ is non-nil, message is flaged as read.")
 
 (luna-define-generic elmo-message-fetch (folder number strategy
 						&optional
-						section
-						outbuf
-						unread)
-  "Fetch a message and return as a string.
+						unread
+						section)
+  "Fetch a message into current buffer.
 FOLDER is the ELMO folder structure.
 NUMBER is the number of the message in the FOLDER.
 STRATEGY is the message fetching strategy.
-If optional argument SECTION is specified, only the SECTION of the message
-is fetched (if possible).
-If second optional argument OUTBUF is specified, fetched message is
-inserted to the buffer and returns t if fetch was ended successfully.
-If third optional argument UNREAD is non-nil, message is not flaged as read.
-Returns non-nil if fetching was succeed.")
-
-(luna-define-generic elmo-message-fetch-with-cache-process (folder
-							    number strategy
-							    &optional
-							    section
-							    unread)
-  "Fetch a message into current buffer with cache process.
-FOLDER is the ELMO folder structure.
-NUMBER is the number of the message in the FOLDER.
-STRATEGY is the message fetching strategy.
-If optional argument SECTION is specified, only the SECTION of the message
-is fetched (if possible).
-If second optional argument UNREAD is non-nil, message is not flaged as read.
+If optional argument UNREAD is non-nil, message is not flaged as read.
+If second optional argument SECTION is specified, only the
+SECTION of the message is fetched (if possible).
 Returns non-nil if fetching was succeed.")
 
 (luna-define-generic elmo-message-fetch-internal (folder number strategy
@@ -696,16 +733,17 @@ Return a cons cell of (NUMBER-CROSSPOSTS . NEW-FLAG-ALIST).")
   (elmo-generic-folder-commit folder))
 
 (defun elmo-generic-folder-commit (folder)
-  (let ((msgdb (elmo-folder-msgdb-internal folder)))
-    (when (and msgdb (elmo-folder-persistent-p folder))
-      (when (elmo-msgdb-message-modified-p msgdb)
-	(elmo-folder-set-info-max-by-numdb
-	 folder
-	 (elmo-folder-list-messages folder nil 'in-msgdb))
-	(elmo-msgdb-killed-list-save
-	 (elmo-folder-msgdb-path folder)
-	 (elmo-folder-killed-list-internal folder)))
-      (elmo-msgdb-save msgdb))))
+  (when (elmo-folder-persistent-p folder)
+    (let ((msgdb (elmo-folder-msgdb-internal folder)))
+      (when msgdb
+	(when (elmo-msgdb-message-modified-p msgdb)
+	  (elmo-folder-set-info-max-by-numdb
+	   folder
+	   (elmo-folder-list-messages folder nil 'in-msgdb)))
+	(elmo-msgdb-save msgdb)))
+    (elmo-msgdb-killed-list-save
+     (elmo-folder-msgdb-path folder)
+     (elmo-folder-killed-list-internal folder))))
 
 (luna-define-method elmo-folder-close-internal ((folder elmo-folder))
   ;; do nothing.
@@ -746,16 +784,24 @@ Return a cons cell of (NUMBER-CROSSPOSTS . NEW-FLAG-ALIST).")
     t))
 
 (luna-define-method elmo-folder-rename ((folder elmo-folder) new-name)
-  (let* ((new-folder (elmo-make-folder new-name)))
+  (let* ((new-folder (elmo-make-folder
+		      new-name
+		      nil
+		      (elmo-folder-mime-charset-internal folder))))
     (unless (eq (elmo-folder-type-internal folder)
 		(elmo-folder-type-internal new-folder))
       (error "Not same folder type"))
-    (if (or (file-exists-p (elmo-folder-msgdb-path new-folder))
-	    (elmo-folder-exists-p new-folder))
-	(error "Already exists folder: %s" new-name))
+    (when (or (file-exists-p (elmo-folder-msgdb-path new-folder))
+	      (elmo-folder-exists-p new-folder))
+      (error "Already exists folder: %s" new-name))
     (elmo-folder-send folder 'elmo-folder-rename-internal new-folder)
     (elmo-msgdb-rename-path folder new-folder)))
 
+(luna-define-method elmo-folder-delete-messages ((folder elmo-folder)
+						 numbers)
+  (and (elmo-folder-delete-messages-internal folder numbers)
+       (elmo-folder-detach-messages folder numbers)))
+
 (luna-define-method elmo-folder-search ((folder elmo-folder)
 					condition
 					&optional numbers)
@@ -765,6 +811,14 @@ Return a cons cell of (NUMBER-CROSSPOSTS . NEW-FLAG-ALIST).")
     (setq results (elmo-msgdb-search msgdb condition numbers))
     (if (listp results)
 	results
+      (elmo-condition-optimize condition)
+      (when (and (consp condition)
+		 (eq (car condition) 'and)
+		 (listp (setq results (elmo-msgdb-search msgdb
+							 (nth 1 condition)
+							 numbers))))
+	(setq numbers results
+	      condition (nth 2 condition)))
       (let ((len (length numbers))
 	    matched)
 	(elmo-with-progress-display (> len elmo-display-progress-threshold)
@@ -786,6 +840,22 @@ Return a cons cell of (NUMBER-CROSSPOSTS . NEW-FLAG-ALIST).")
 	(message "Searching...done")
 	(nreverse matched)))))
 
+(defun elmo-message-buffer-match-condition (condition number)
+  (let* ((handler (luna-make-entity 'modb-buffer-entity-handler))
+	 (result (elmo-condition-match
+		  condition
+		  (lambda (condition handler entity)
+		    (elmo-msgdb-message-match-condition handler
+							condition
+							entity))
+		  (list
+		   handler
+		   (elmo-msgdb-make-message-entity
+		    handler
+		    :number number
+		    :buffer (current-buffer))))))
+    (and result (not (elmo-filter-condition-p result)))))
+
 (luna-define-method elmo-message-match-condition ((folder elmo-folder)
 						  number condition
 						  numbers)
@@ -809,11 +879,11 @@ Return a cons cell of (NUMBER-CROSSPOSTS . NEW-FLAG-ALIST).")
 						      (and cache t)
 						      nil
 						      cache-path)
-			    nil (current-buffer) t)
+			    'unread)
 	(set-buffer-multibyte default-enable-multibyte-characters)
 	(decode-coding-region (point-min) (point-max)
 			      elmo-mime-display-as-is-coding-system)
-	(elmo-buffer-field-condition-match condition number numbers)))))
+	(elmo-message-buffer-match-condition condition number)))))
 
 (luna-define-method elmo-folder-pack-numbers ((folder elmo-folder))
   nil) ; default is noop.
@@ -943,13 +1013,12 @@ If optional argument IF-EXISTS is nil, load on demand.
 
 (defsubst elmo-strict-folder-diff (folder)
   "Return folder diff information strictly from FOLDER."
-  (let ((in-db (sort (elmo-msgdb-list-messages (elmo-folder-msgdb folder))
-		     '<))
+  (let ((in-db (sort (elmo-folder-list-messages folder nil 'in-msgdb) '<))
 	(in-folder  (elmo-folder-list-messages folder))
 	append-list delete-list diff)
     (cons (if (equal in-folder in-db)
 	      0
-	    (setq diff (elmo-list-diff in-folder in-db nil))
+	    (setq diff (elmo-list-diff in-folder in-db))
 	    (setq append-list (car diff))
 	    (setq delete-list (cadr diff))
 	    (if append-list
@@ -1047,15 +1116,11 @@ If optional argument IF-EXISTS is nil, load on demand.
 				 'entire t nil
 				 (elmo-file-cache-path cache)))
 			   (error "Unplugged")))
-		     nil (current-buffer)
 		     'unread)
 		    (> (buffer-size) 0)
 		    (elmo-folder-append-buffer
 		     folder
-		     (elmo-message-flags
-		      src-folder
-		      (car numbers)
-		      (elmo-msgdb-get-message-id-from-buffer))
+		     (elmo-message-flags-for-append src-folder (car numbers))
 		     (if same-number (car numbers))))))
 	  (error (setq failure t)))
 	;; FETCH & APPEND finished
@@ -1094,8 +1159,7 @@ If optional argument IF-EXISTS is nil, load on demand.
 	  (elmo-folder-close dst-folder)))
       (if (and (not no-delete) succeeds)
 	  (progn
-	    (if (and (elmo-folder-delete-messages src-folder succeeds)
-		     (elmo-folder-detach-messages src-folder succeeds))
+	    (if (elmo-folder-delete-messages src-folder succeeds)
 		(progn
 		  (elmo-global-flag-detach-messages
 		   src-folder succeeds (eq dst-folder 'null))
@@ -1119,7 +1183,19 @@ If optional argument IF-EXISTS is nil, load on demand.
   (or (elmo-folder-path-internal folder)
       (elmo-folder-set-path-internal
        folder
-       (elmo-folder-expand-msgdb-path folder))))
+       (if (null elmo-msgdb-path-encode-threshold)
+	   (elmo-folder-expand-msgdb-path folder)
+	 (let* ((path (directory-file-name
+		       (elmo-folder-expand-msgdb-path folder)))
+		(dirname (file-name-nondirectory path)))
+	   (if (<= (length dirname) elmo-msgdb-path-encode-threshold)
+	       path
+	     (require 'md5)
+	     (setq dirname (md5 dirname))
+	     (when (> (length dirname) elmo-msgdb-path-encode-threshold)
+	       (error "Cannot shrink msgdb path for `%s'"
+		      (elmo-folder-name-internal folder)))
+	     (expand-file-name dirname (file-name-directory path))))))))
 
 (luna-define-generic elmo-message-cached-p (folder number)
   "Return non-nil if the message is cached.")
@@ -1132,6 +1208,7 @@ If optional argument IF-EXISTS is nil, load on demand.
 Return non-nil when message is accessible."
   (or (elmo-folder-plugged-p folder)
       (elmo-folder-local-p folder)
+      (< number 0) ; in dop spool
       (elmo-message-cached-p folder number)))
 
 (luna-define-generic elmo-message-set-cached (folder number cached)
@@ -1144,12 +1221,19 @@ If CACHED is t, message is set as cached.")
 					     number cached)
   (if cached
       (elmo-msgdb-set-flag (elmo-folder-msgdb folder) number 'cached)
-    (elmo-msgdb-unset-flag (elmo-folder-msgdb folder) number 'cached)))
+    (elmo-msgdb-unset-flag (elmo-folder-msgdb folder) number 'cached))
+  (elmo-emit-signal 'cache-changed folder number))
 
 (defun elmo-message-copy-entity (entity)
   (elmo-msgdb-copy-message-entity (elmo-message-entity-handler entity)
 				  entity))
 
+(luna-define-generic elmo-message-number (folder message-id)
+  "Get message number from MSGDB which corresponds to MESSAGE-ID.")
+
+(luna-define-method elmo-message-number ((folder elmo-folder) message-id)
+  (elmo-msgdb-message-number (elmo-folder-msgdb folder) message-id))
+
 (luna-define-generic elmo-message-entity (folder key)
   "Return the message-entity structure which matches to the KEY.
 KEY is a number or a string.
@@ -1186,14 +1270,14 @@ ENTITY is the message-entity to get the parent.")
 
 (luna-define-generic elmo-folder-count-flags (folder)
   "Count flagged message number in the msgdb of the FOLDER.
-Return a list of numbers (`new' `unread' `answered')")
+Return alist of flag and numbers.
+Example:
+\(\(new . 10\)
+  \(unread . 20\)
+  \(answered . 3\)\)")
 
 (luna-define-method elmo-folder-count-flags ((folder elmo-folder))
-  (let* ((flag-count (elmo-msgdb-flag-count (elmo-folder-msgdb folder)))
-	 (new (or (cdr (assq 'new flag-count)) 0))
-	 (unread (or (cdr (assq 'unread flag-count)) 0))
-	 (answered(or (cdr (assq 'answered flag-count)) 0)))
-    (list new (- unread new) answered)))
+  (elmo-msgdb-flag-count (elmo-folder-msgdb folder)))
 
 (defun elmo-message-set-flag (folder number flag &optional is-local)
   "Set message flag.
@@ -1225,14 +1309,16 @@ If optional IS-LOCAL is non-nil, update only local (not server) status."
   ;; XXX Transitional implementation.
   (elmo-folder-unset-flag folder (list number) flag is-local))
 
-(luna-define-generic elmo-message-field (folder number field)
+(luna-define-generic elmo-message-field (folder number field &optional type)
   "Get message field value in the msgdb.
 FOLDER is the ELMO folder structure.
 NUMBER is a number of the message.
-FIELD is a symbol of the field.")
+FIELD is a symbol of the field.
+If optional argument TYPE is specified, return converted value.")
 
-(luna-define-method elmo-message-field ((folder elmo-folder) number field)
-  (elmo-message-entity-field (elmo-message-entity folder number) field))
+(luna-define-method elmo-message-field ((folder elmo-folder)
+					number field &optional type)
+  (elmo-msgdb-message-field (elmo-folder-msgdb folder) number field type))
 
 (luna-define-generic elmo-message-set-field (folder number field value)
   "Set message field value in the msgdb.
@@ -1258,12 +1344,17 @@ VALUE is a value to set.")
 					  &optional is-local)
   (when (elmo-folder-msgdb-internal folder)
     (dolist (number numbers)
-      (when (elmo-global-flag-p flag)
-	(let ((message-id (elmo-message-field folder number 'message-id)))
-	  (elmo-global-flag-set flag folder number message-id)))
-      (elmo-msgdb-set-flag (elmo-folder-msgdb folder)
-			   number
-			   flag))))
+      (let ((old-flags (elmo-message-flags folder number)))
+	(when (elmo-global-flag-p flag)
+	  (let ((message-id (elmo-message-field folder number 'message-id)))
+	    (elmo-global-flag-set flag folder number message-id)))
+	(elmo-msgdb-set-flag (elmo-folder-msgdb folder) number flag)
+	(elmo-emit-signal 'flag-changing
+			  folder
+			  number
+			  old-flags
+			  (elmo-message-flags folder number))))
+    (elmo-emit-signal 'flag-changed folder numbers)))
 
 (defun elmo-message-has-global-flag-p (folder number)
   "Return non-nil when the message in the FOLDER with NUMBER has global flag."
@@ -1297,11 +1388,16 @@ If Optional LOCAL is non-nil, don't update server flag."
 					    &optional is-local)
   (when (elmo-folder-msgdb-internal folder)
     (dolist (number numbers)
-      (when (elmo-global-flag-p flag)
-	(elmo-global-flag-detach flag folder number 'always))
-      (elmo-msgdb-unset-flag (elmo-folder-msgdb folder)
-			     number
-			     flag))))
+      (let ((old-flags (elmo-message-flags folder number)))
+	(when (elmo-global-flag-p flag)
+	  (elmo-global-flag-detach flag folder number 'always))
+	(elmo-msgdb-unset-flag (elmo-folder-msgdb folder) number flag)
+	(elmo-emit-signal 'flag-changing
+			  folder
+			  number
+			  old-flags
+			  (elmo-message-flags folder number))))
+    (elmo-emit-signal 'flag-changed folder numbers)))
 
 (luna-define-method elmo-folder-process-crosspost ((folder elmo-folder))
   ;; Do nothing.
@@ -1329,11 +1425,11 @@ If Optional LOCAL is non-nil, don't update server flag."
 	       ;; Let duplicates be a temporary killed message.
 	       (elmo-folder-kill-messages folder duplicates)
 	       ;; Should be flag as read.
-	       (elmo-folder-set-flag folder duplicates 'read))
+	       (elmo-folder-unset-flag folder duplicates 'unread))
 	      ((eq (elmo-folder-process-duplicates-internal folder)
 		   'read)
 	       ;; Flag as read duplicates.
-	       (elmo-folder-set-flag folder duplicates 'read))
+	       (elmo-folder-unset-flag folder duplicates 'unread))
 	      (t
 	       ;; Do nothing.
 	       (setq duplicates nil)))
@@ -1371,26 +1467,21 @@ If Optional LOCAL is non-nil, don't update server flag."
 						      number strategy)
   nil)
 
+(defun elmo-message-fetch-string (folder number strategy
+					 &optional
+					 unread
+					 section)
+  (with-temp-buffer
+    (set-buffer-multibyte nil)
+    (when (elmo-message-fetch folder number strategy unread section)
+      (buffer-string))))
+
 (luna-define-method elmo-message-fetch ((folder elmo-folder)
 					number strategy
 					&optional
-					section
-					outbuf
-					unread)
-  (if outbuf
-      (with-current-buffer outbuf
-	(erase-buffer)
-	(elmo-message-fetch-with-cache-process folder number
-					       strategy section unread))
-    (with-temp-buffer
-      (elmo-message-fetch-with-cache-process folder number
-					     strategy section unread)
-      (buffer-string))))
-
-(luna-define-method elmo-message-fetch-with-cache-process ((folder elmo-folder)
-							   number strategy
-							   &optional
-							   section unread)
+					unread
+					section)
+  (erase-buffer)
   (let ((cache-path (elmo-fetch-strategy-cache-path strategy))
 	(method-priorities
 	 (cond ((eq (elmo-fetch-strategy-use-cache strategy) 'maybe)
@@ -1399,28 +1490,33 @@ If Optional LOCAL is non-nil, don't update server flag."
 		'(cache entity))
 	       (t
 		'(entity))))
-	result err)
+	result err updated-server-flag)
     (while (and method-priorities
-		(null result))
+		(not result))
       (setq result
 	    (case (car method-priorities)
 	      (cache
 	       (elmo-file-cache-load cache-path section))
 	      (entity
-	       (when (and (condition-case error
-			      (elmo-message-fetch-internal folder number
-							   strategy
-							   section
-							   unread)
-			    (error (setq err error) nil))
-			  (> (buffer-size) 0))
+	       (when (condition-case error
+			 (elmo-message-fetch-internal folder number
+						      strategy
+						      section
+						      unread)
+		       (error (setq err error) nil))
+		 (setq updated-server-flag t)
 		 (when (and (elmo-fetch-strategy-save-cache strategy)
 			    cache-path)
 		   (elmo-file-cache-save cache-path section))
 		 t)))
 	    method-priorities (cdr method-priorities)))
-    (or result
-	(and err (signal (car err) (cdr err))))))
+    (if result
+	(when (and (not unread)
+		   (elmo-message-flagged-p folder number 'unread))
+	  (elmo-message-unset-flag folder number 'unread updated-server-flag))
+      (when err
+	(signal (car err) (cdr err))))
+    result))
 
 (defun elmo-folder-kill-messages-range (folder beg end)
   (elmo-folder-set-killed-list-internal
@@ -1472,48 +1568,46 @@ If update process is interrupted, return nil.")
 					     ignore-msgdb
 					     no-check
 					     mask)
-  (let ((killed-list (elmo-folder-killed-list-internal folder))
-	(before-append t)
-	old-msgdb diff diff-2 delete-list new-list new-msgdb flag
-	flag-table crossed after-append)
-    (setq old-msgdb (elmo-folder-msgdb folder))
-    (setq flag-table (elmo-flag-table-load (elmo-folder-msgdb-path folder)))
+  (let ((old-msgdb (elmo-folder-msgdb folder))
+	(killed-list (elmo-folder-killed-list-internal folder))
+	(flag-table (elmo-flag-table-load (elmo-folder-msgdb-path folder)))
+	(before-append t))
     (when ignore-msgdb
       (elmo-msgdb-flag-table (elmo-folder-msgdb folder) flag-table)
       (elmo-folder-clear folder (not disable-killed)))
     (unless no-check (elmo-folder-check folder))
     (condition-case nil
-	(progn
+	(let ((killed-list (elmo-folder-killed-list-internal folder))
+	      diff-new diff-del
+	      delete-list new-list new-msgdb crossed)
 	  (message "Checking folder diff...")
-	  (setq diff (elmo-list-diff (elmo-folder-list-messages
-				      folder
-				      (not disable-killed))
-				     (elmo-folder-list-messages
-				      folder
-				      (not disable-killed)
-				      'in-msgdb)))
-	  (when (and mask (car diff))
-	    (setcar diff (elmo-list-filter mask (car diff))))
+	  (elmo-set-list
+	   '(diff-new diff-del)
+	   (elmo-list-diff (elmo-folder-list-messages folder)
+			   (elmo-folder-list-messages folder nil 'in-msgdb)))
+	  (when diff-new
+	    (setq diff-new (sort diff-new #'<))
+	    (unless disable-killed
+	      (setq diff-new (elmo-living-messages diff-new killed-list)))
+	    (when (and mask (not ignore-msgdb))
+	      (setq diff-new (elmo-list-filter mask diff-new))))
 	  (message "Checking folder diff...done")
-	  (setq new-list (elmo-folder-confirm-appends (car diff)))
-	  ;; Set killed list as ((1 . MAX-OF-DISAPPEARED))
-	  (when (and (not (eq (length (car diff))
-			      (length new-list)))
-		     (setq diff-2 (elmo-list-diff (car diff) new-list)))
-	    (elmo-folder-kill-messages-range
-	     folder
-	     (car (car diff-2))
-	     (nth (- (length (car diff-2)) 1) (car diff-2))))
-	  (setq delete-list (cadr diff))
-	  (if (or (equal diff '(nil nil))
-		  (equal diff '(nil))
-		  (and (eq (length (car diff)) 0)
-		       (eq (length (cadr diff)) 0)))
+	  (setq new-list (elmo-folder-confirm-appends diff-new))
+	  ;; Append to killed list as (MIN-OF-DISAPPEARED . MAX-OF-DISAPPEARED)
+	  (when (not (eq (length diff-new)
+			 (length new-list)))
+	    (let* ((diff (elmo-list-diff diff-new new-list))
+		   (disappeared (sort (car diff) #'<)))
+	      (when disappeared
+		(elmo-folder-kill-messages-range folder
+						 (car disappeared)
+						 (elmo-last disappeared)))))
+	  (setq delete-list diff-del)
+	  (if (and (null diff-new) (null diff-del))
 	      (progn
 		(elmo-folder-update-number folder)
 		(elmo-folder-process-crosspost folder)
-		0 ; no updates.
-		)
+		0)			; `0' means no updates.
 	    (when delete-list
 	      (elmo-folder-detach-messages folder delete-list))
 	    (when new-list
@@ -1542,7 +1636,14 @@ If update process is interrupted, return nil.")
 
 (luna-define-method elmo-folder-detach-messages ((folder elmo-folder)
 						 numbers)
-  (elmo-msgdb-delete-messages (elmo-folder-msgdb folder) numbers))
+  (when (elmo-msgdb-delete-messages (elmo-folder-msgdb folder) numbers)
+    ;; Remove NUMBERS from killed message list.
+    (elmo-folder-set-killed-list-internal
+     folder
+     (elmo-number-set-delete-list
+      (elmo-folder-killed-list-internal folder)
+      numbers))
+    t))
 
 (luna-define-generic elmo-folder-length (folder)
   "Return number of messages in the FOLDER.")
@@ -1555,7 +1656,8 @@ If update process is interrupted, return nil.")
 (defun elmo-folder-msgdb-load (folder &optional silent)
   (unless silent
     (message "Loading msgdb for %s..." (elmo-folder-name-internal folder)))
-  (let ((msgdb (elmo-load-msgdb (elmo-folder-msgdb-path folder))))
+  (let ((msgdb (elmo-load-msgdb (elmo-folder-msgdb-path folder)
+				(elmo-folder-mime-charset-internal folder))))
     (elmo-folder-set-info-max-by-numdb
      folder
      (elmo-msgdb-list-messages msgdb))
@@ -1627,6 +1729,7 @@ Return a hashtable for newsgroups."
     (elmo-make-directory temp-dir)
     temp-dir))
 
+;;;
 (defun elmo-init ()
   "Initialize ELMO module."
   (elmo-crosspost-message-alist-load)
@@ -1707,6 +1810,8 @@ Return a hashtable for newsgroups."
 			       'elmo-cache-directory)
 (elmo-define-obsolete-variable 'elmo-msgdb-dir
 			       'elmo-msgdb-directory)
+(elmo-define-obsolete-variable 'elmo-global-flag-list
+			       'elmo-global-flags)
 
 ;; Obsolete functions.
 ;; 2001-12-11: *-dir -> *-directory
