@@ -449,35 +449,31 @@ Return number if put mark succeed"
     (let ((start (point))
 	  (failures 0)
 	  (refile-len (length mark-list))
-	  dst-msgs ; loop counter
-	  result)
+	  dst-msgs)
       ;; begin refile...
-      (setq dst-msgs
-	    (wl-summary-make-destination-numbers-list mark-list))
+      (setq dst-msgs (wl-summary-make-destination-numbers-list mark-list))
       (goto-char start)	; avoid moving cursor to the bottom line.
       (when (> refile-len elmo-display-progress-threshold)
 	(elmo-progress-set 'elmo-folder-move-messages
 			   refile-len "Refiling messages..."))
-      (while dst-msgs
-	(setq result nil)
-	(condition-case nil
-	    (setq result (elmo-folder-move-messages
-			  wl-summary-buffer-elmo-folder
-			  (cdr (car dst-msgs))
-			  (wl-folder-get-elmo-folder (car (car dst-msgs)))))
-	  (error nil))
-	(if result		; succeeded.
+      (dolist (pair dst-msgs)
+	(if (condition-case nil
+		(elmo-folder-move-messages
+		 wl-summary-buffer-elmo-folder
+		 (cdr pair)
+		 (wl-folder-get-elmo-folder (car pair)))
+	      (error nil))
 	    (progn
 	      ;; update buffer.
-	      (wl-summary-delete-messages-on-buffer (cdr (car dst-msgs)))
+	      (wl-summary-delete-messages-on-buffer (cdr pair))
 	      (setq wl-summary-buffer-temp-mark-list
 		    (wl-delete-associations
-		     (cdr (car dst-msgs))
+		     (cdr pair)
 		     wl-summary-buffer-temp-mark-list)))
-	  (setq failures
-		(+ failures (length (cdr (car dst-msgs))))))
-	(setq dst-msgs (cdr dst-msgs)))
+	  (setq failures (+ failures (length (cdr pair))))))
       (elmo-progress-clear 'elmo-folder-move-messages)
+      (if (<= failures 0)
+	  (message "Refiling messages...done"))
       failures)))
 
 ;; Copy action
@@ -489,8 +485,7 @@ Return number if put mark succeed"
     (let ((start (point))
 	  (failures 0)
 	  (refile-len (length mark-list))
-	  dst-msgs ; loop counter
-	  result)
+	  dst-msgs)
       ;; begin refile...
       (setq dst-msgs
 	    (wl-summary-make-destination-numbers-list mark-list))
@@ -498,27 +493,25 @@ Return number if put mark succeed"
       (when (> refile-len elmo-display-progress-threshold)
 	(elmo-progress-set 'elmo-folder-move-messages
 			   refile-len "Copying messages..."))
-      (while dst-msgs
-	(setq result nil)
-	(condition-case nil
-	    (setq result (elmo-folder-move-messages
-			  wl-summary-buffer-elmo-folder
-			  (cdr (car dst-msgs))
-			  (wl-folder-get-elmo-folder (car (car dst-msgs)))
-			  'no-delete))
-	  (error nil))
-	(if result		; succeeded.
+      (dolist (pair dst-msgs)
+	(if (condition-case nil
+		(elmo-folder-move-messages
+		 wl-summary-buffer-elmo-folder
+		 (cdr pair)
+		 (wl-folder-get-elmo-folder (car pair))
+		 'no-delete)
+	      (error nil))
 	    (progn
 	      ;; update buffer.
-	      (wl-summary-delete-copy-marks-on-buffer (cdr (car dst-msgs)))
+	      (wl-summary-delete-copy-marks-on-buffer (cdr pair))
 	      (setq wl-summary-buffer-temp-mark-list
-		    (wl-delete-associations 
-		     (cdr (car dst-msgs))
+		    (wl-delete-associations
+		     (cdr pair)
 		     wl-summary-buffer-temp-mark-list)))
-	  (setq failures
-		(+ failures (length (cdr (car dst-msgs))))))
-	(setq dst-msgs (cdr dst-msgs)))
+	  (setq failures (+ failures (length (cdr pair))))))
       (elmo-progress-clear 'elmo-folder-move-messages)
+      (if (<= failures 0)
+	  (message "Copying messages...done"))
       failures)))
 
 ;; Prefetch.
@@ -543,7 +536,7 @@ Return number if put mark succeed"
 ;; Resend.
 (defun wl-summary-get-resend-address (action number)
   "Decide resend address."
-  (completing-read "Resend message to: " 'wl-complete-address))
+  (wl-address-read-from-minibuffer "Resend message to: "))
 
 (defun wl-summary-exec-action-resend (mark-list)
   (let ((failure 0))
@@ -565,7 +558,7 @@ Return number if put mark succeed"
     (save-excursion
       ;; We first set up a normal mail buffer.
       (set-buffer (get-buffer-create " *wl-draft-resend*"))
-      (buffer-disable-undo (current-buffer))
+      (set-buffer-multibyte nil)
       (erase-buffer)
       (setq wl-sent-message-via nil)
       ;; Insert our usual headers.
@@ -583,11 +576,16 @@ Return number if put mark succeed"
       (let ((beg (point)))
 	;; Insert the message to be resent.
 	(insert
-	 (with-temp-buffer
-	   (elmo-message-fetch folder number
-			       (elmo-make-fetch-strategy 'entire)
-			       nil (current-buffer) 'unread)
-	   (buffer-string)))
+	 ;; elmo-message-fetch is erase current buffer before fetch message
+	 (elmo-message-fetch-string folder number
+				    (if wl-summary-resend-use-cache
+					(elmo-make-fetch-strategy
+					 'entire 'maybe nil
+					 (elmo-file-cache-get-path
+					  (elmo-message-field
+					   folder number 'message-id)))
+				      (elmo-make-fetch-strategy 'entire))
+				    'unread))
 	(goto-char (point-min))
 	(search-forward "\n\n")
 	(forward-char -1)
@@ -604,6 +602,7 @@ Return number if put mark succeed"
 	(goto-char beg)
 	(when (looking-at "From ")
 	  (replace-match "X-From-Line: ")))
+      (run-hooks 'wl-summary-resend-hook)
       ;; Send it.
       (wl-draft-dispatch-message)
       (kill-buffer (current-buffer)))
@@ -659,7 +658,7 @@ Return number if put mark succeed"
   (let ((failures 0)
 	collected pair action modified)
     (dolist (action wl-summary-mark-action-list)
-      (setq collected (cons (cons 
+      (setq collected (cons (cons
 			     (wl-summary-action-mark action)
 			     nil) collected)))
     (dolist (mark-info wl-summary-buffer-temp-mark-list)
@@ -672,12 +671,13 @@ Return number if put mark succeed"
     ;; collected is a pair of
     ;; mark-string and a list of mark-info
     (dolist (pair collected)
-      (setq action (assoc (car pair) wl-summary-mark-action-list))
-      (when (and (cdr pair) (wl-summary-action-exec-function action))
-	(setq modified t)
-	(setq failures (+ failures (funcall
-				    (wl-summary-action-exec-function action)
-				    (cdr pair))))))
+      (when (cdr pair)
+	(setq action (assoc (car pair) wl-summary-mark-action-list))
+	(when (wl-summary-action-exec-function action)
+	  (setq modified t)
+	  (setq failures (+ failures (funcall
+				      (wl-summary-action-exec-function action)
+				      (cdr pair)))))))
     (when modified
       (wl-summary-set-message-modified))
     (run-hooks 'wl-summary-exec-hook)
@@ -847,8 +847,9 @@ Return number if put mark succeed"
 					       (wl-thread-get-entity number))))
 				   (wl-thread-entity-get-descendant
 				    thr-entity))))
-	  (when (and (not (wl-summary-no-auto-refile-message-p
-			   number))
+	  (when (and (not (wl-summary-no-auto-refile-message-p number))
+		     (not (wl-summary-reserve-temp-mark-p
+			   (nth 1 (wl-summary-registered-temp-mark number))))
 		     (setq dst
 			   (wl-folder-get-realname
 			    (wl-refile-guess
