@@ -441,6 +441,7 @@
     (wl-folder-nmz-image	  . wl-nmz-folder-icon)
     (wl-folder-shimbun-image      . wl-shimbun-folder-icon)
     (wl-folder-file-image	  . wl-file-folder-icon)
+    (wl-folder-access-image	  . wl-access-folder-icon)
     (wl-folder-trash-empty-image  . wl-empty-trash-folder-icon)
     (wl-folder-draft-image	  . wl-draft-folder-icon)
     (wl-folder-queue-image	  . wl-queue-folder-icon)
@@ -534,6 +535,7 @@
     (define-key keymap "l" 'wl-message-toggle-disp-summary)
     (define-key keymap "\C-c:d" 'wl-message-decrypt-pgp-nonmime)
     (define-key keymap "\C-c:v" 'wl-message-verify-pgp-nonmime)
+    (define-key keymap "w" 'wl-draft)
     (define-key keymap [mouse-4] 'wl-message-wheel-down)
     (define-key keymap [mouse-5] 'wl-message-wheel-up)
     (define-key keymap [S-mouse-4] 'wl-message-wheel-down)
@@ -631,7 +633,8 @@
 See info under Wanderlust for full documentation.
 
 Special commands:
-\\{wl-draft-mode-map}"))
+\\{wl-draft-mode-map}"
+    (setq font-lock-defaults nil)))
 
 (defun wl-draft-key-setup ()
   (define-key wl-draft-mode-map "\C-c\C-y" 'wl-draft-yank-original)
