@@ -182,14 +182,24 @@ See `wl-summary-mark-action-list' for the detail of element."
   (let ((folder (or folder wl-summary-buffer-elmo-folder))
 	(number (or number (wl-summary-message-number)))
 	spam)
-    (message "Cheking spam...")
+    (message "Checking spam...")
     (when (setq spam (elmo-spam-message-spam-p (elmo-spam-processor)
 					       folder number))
       (wl-summary-spam number))
-    (message "Cheking spam...done")
+    (message "Checking spam...done")
     (when (interactive-p)
       (message "No: %d is %sa spam message." number (if spam "" "not ")))))
 
+(defun wl-summary-test-spam-region (beg end)
+  (interactive "r")
+  (let ((numbers (wl-summary-collect-numbers-region beg end)))
+    (cond (numbers
+	   (wl-spam-map-spam-messages wl-summary-buffer-elmo-folder
+				      numbers
+				      #'wl-summary-spam))
+	  ((interactive-p)
+	   (message "No message to test.")))))
+
 (defun wl-summary-mark-spam (&optional all)
   "Set spam mark to messages which is spam classification."
   (interactive "P")
@@ -318,11 +328,11 @@ See `wl-summary-mark-action-list' for the detail of element."
   (let ((original (wl-message-get-original-buffer))
 	(number wl-message-buffer-cur-number)
 	spam)
-    (message "Cheking spam...")
+    (message "Checking spam...")
     (when (setq spam (elmo-spam-buffer-spam-p (elmo-spam-processor) original))
       (with-current-buffer wl-message-buffer-cur-summary-buffer
 	(wl-summary-spam number)))
-    (message "Cheking spam...done")
+    (message "Checking spam...done")
     (message "No: %d is %sa spam message." number (if spam "" "not "))))
 
 (defun wl-refile-guess-by-spam (entity)
@@ -332,7 +342,7 @@ See `wl-summary-mark-action-list' for the detail of element."
     wl-spam-folder))
 
 (defun wl-spam-setup ()
-  (add-hook 'wl-summary-prepared-hook #'wl-summary-auto-check-spam)
+  (add-hook 'wl-summary-sync-updated-hook #'wl-summary-auto-check-spam)
   (let ((actions wl-summary-mark-action-list)
 	action)
     (while actions
@@ -353,6 +363,8 @@ See `wl-summary-mark-action-list' for the detail of element."
 	    (cons (wl-summary-action-mark action)
 		  wl-summary-skip-mark-list))))
   (define-key wl-summary-mode-map "k" wl-summary-spam-map)
+  (define-key
+    wl-summary-mode-map "rkc" 'wl-summary-test-spam-region)
   (define-key
     wl-summary-mode-map "mk" 'wl-summary-target-mark-spam)
   (define-key
