@@ -119,30 +119,22 @@
 (defun wl-thread-save-top-list (dir)
   (let ((top-file (expand-file-name wl-thread-entity-list-file dir))
 	(entity wl-thread-entity-list)
-	(tmp-buffer (get-buffer-create " *wl-thread-save-top-list*"))
 	print-length)
-    (save-excursion
-      (set-buffer tmp-buffer)
-      (erase-buffer)
+    (with-temp-buffer
       (when (file-writable-p top-file)
-	(prin1 entity tmp-buffer)
-	(princ "\n" tmp-buffer)
-	(write-region (point-min) (point-max) top-file nil 'no-msg)
-	(kill-buffer tmp-buffer)))))
+	(prin1 entity (current-buffer))
+	(princ "\n" (current-buffer))
+	(write-region (point-min) (point-max) top-file nil 'no-msg)))))
 
 (defun wl-thread-save-entities (dir)
   (let ((top-file (expand-file-name wl-thread-entity-file dir))
 	(entities wl-thread-entities)
-	(tmp-buffer (get-buffer-create " *wl-thread-save-entities*"))
 	print-length print-level)
-    (save-excursion
-      (set-buffer tmp-buffer)
-      (erase-buffer)
+    (with-temp-buffer
       (when (file-writable-p top-file)
-	(prin1 entities tmp-buffer)
-	(princ "\n" tmp-buffer)
-	(write-region (point-min) (point-max) top-file nil 'no-msg)
-	(kill-buffer tmp-buffer)))))
+	(prin1 entities (current-buffer))
+	(princ "\n" (current-buffer))
+	(write-region (point-min) (point-max) top-file nil 'no-msg)))))
 
 (defsubst wl-thread-entity-get-number (entity)
   (nth 0 entity))
@@ -231,7 +223,7 @@
      (setq entity (wl-thread-get-entity (car msgs))))
    ret-val))
 
-(defsubst wl-thread-entity-get-descendant (entity)
+(defun wl-thread-entity-get-descendant (entity)
   (let (children
 	ret-val msgs-stack
 	(msgs (list (car entity))))
@@ -241,7 +233,7 @@
      (if (null children)
 	 (while (and (null msgs) msgs-stack)
 	   (setq msgs (wl-pop msgs-stack)))
-       (setq ret-val (append ret-val (copy-sequence children)))
+       (setq ret-val (nconc ret-val (copy-sequence children)))
        (wl-push msgs msgs-stack)
        (setq msgs children))
      (setq entity (wl-thread-get-entity (car msgs))))
@@ -375,15 +367,14 @@ ENTITY is returned."
 	 (parent-msg (or parent-msg (wl-thread-entity-get-parent entity)))
 	 (buffer-read-only nil)
 	 (inhibit-read-only t)
-	 message-entity temp-mark summary-line invisible-top dest-pair)
+	 message-entity temp-mark summary-line invisible-top)
     (if (wl-thread-delete-line-from-buffer msg)
 	(progn
 	  (cond
 	   ((memq msg wl-summary-buffer-target-mark-list)
 	    (setq temp-mark "*"))
 	   ((setq temp-mark (wl-summary-registered-temp-mark msg))
-	    (setq dest-pair (cons (nth 0 temp-mark)(nth 2 temp-mark))
-		  temp-mark (nth 1 temp-mark)))
+	    (setq temp-mark (nth 1 temp-mark)))
 	   (t (setq temp-mark (wl-summary-get-score-mark msg))))
 	  (when (setq message-entity
 		      (elmo-message-entity wl-summary-buffer-elmo-folder
@@ -402,10 +393,7 @@ ENTITY is returned."
 		  nil
 		(wl-thread-maybe-get-children-num msg))
 	      (wl-thread-make-indent-string entity)
-	      (wl-thread-entity-get-linked entity)))
-	    (if dest-pair
-		(wl-summary-print-argument (car dest-pair)
-					   (cdr dest-pair)))))
+	      (wl-thread-entity-get-linked entity)))))
       ;; insert thread (moving thread)
       (if (not (setq invisible-top
 		     (wl-thread-entity-parent-invisible-p entity)))
@@ -515,8 +503,14 @@ ENTITY is returned."
     (let ((entity (wl-thread-get-entity msg))
 	  top-child top-entity update-msgs invisible-top)
       (setq wl-summary-buffer-number-list
-	    (delq msg wl-summary-buffer-number-list))	      
+	    (delq msg wl-summary-buffer-number-list))
       (when entity
+	(when deep
+	  (setq wl-summary-buffer-number-list
+		(elmo-list-delete
+		 (wl-thread-entity-get-descendant entity)
+		 wl-summary-buffer-number-list
+		 #'delq)))
 	(let ((parent (wl-thread-entity-get-parent-entity entity)))
 	  (if parent
 	      ;; has parent.
@@ -1073,7 +1067,7 @@ Message is inserted to the summary buffer."
       (incf depth)
       (setq entity (wl-thread-get-entity number)))
     depth))
-  
+
 (defun wl-thread-update-indent-string-region (beg end)
   (interactive "r")
   (save-excursion
@@ -1144,7 +1138,15 @@ Message is inserted to the summary buffer."
 	    (wl-thread-entity-set-children
 	     dst-parent-entity
 	     (append children (list number)))
-	    (wl-thread-entity-set-linked entity t))
+	    (wl-thread-entity-set-linked
+	     entity
+	     (let ((parent (elmo-message-entity-parent
+			    wl-summary-buffer-elmo-folder
+			    (elmo-message-entity
+			     wl-summary-buffer-elmo-folder
+			     number))))
+	       (or (null parent)
+		   (/= parent-number (elmo-message-entity-number parent))))))
 	;; insert as top
 	(wl-append wl-thread-entity-list (list number))
 	(wl-thread-entity-set-linked entity nil))
